from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import Awaitable, Callable, Iterator
from contextlib import contextmanager
from datetime import datetime
from functools import partial
import importlib.resources
import os
from types import SimpleNamespace
import typing
from typing import Any, TypeVar

import aiohttp
import aiohttp.web
from aiohttp_rpc import JsonRpcMethod
from aiohttp_rpc import middlewares as rpc_middlewares
from aiohttp_rpc.errors import InvalidParams as InvalidParamsError
from aiohttp_rpc.errors import ServerError
from aiohttp_rpc.server import WsJsonRpcServer
import click
from loguru import logger
from pydantic import BaseModel, ValidationError
import sqlalchemy as sa
from typing_extensions import Concatenate, Literal, ParamSpec, TypeAlias, TypedDict
from yarl import URL

from instawow import __version__, db, matchers, models
from instawow import results as R
from instawow.common import Strategy
from instawow.config import Config
from instawow.manager import Manager, TraceRequestCtx, init_web_client, is_outdated
from instawow.resolvers import CatalogueEntry, Defn
from instawow.utils import run_in_thread as t
from instawow.utils import uniq

from . import frontend, templates

_T = TypeVar('_T')
_P = ParamSpec('_P')
_ManagerWorkQueueItem: TypeAlias = 'tuple[asyncio.Future[Any], str, Callable[..., Awaitable[Any]]]'


LOCALHOST = '127.0.0.1'


class _ConfigError(ServerError):
    code = -32001
    message = 'invalid configuration parameters'


@contextmanager
def _reraise_validation_error(
    error_class: type[ServerError | InvalidParamsError] = ServerError,
    values: dict[Any, Any] | None = None,
) -> Iterator[None]:
    try:
        yield
    except ValidationError as error:
        errors = error.errors()
        logger.info(f'received invalid request: {(values, errors)}')
        raise error_class(data=errors) from error


methods: list[tuple[str, type[BaseParams]]] = []


def _register_method(method: str):
    def inner(param_class: type[BaseParams]):
        methods.append((method, param_class))
        return param_class

    return inner


class BaseParams(BaseModel):
    @classmethod
    def bind(cls, method: str, managers: _ManagerWorkQueue) -> JsonRpcMethod:
        async def respond(**kwargs: Any) -> BaseModel:
            with _reraise_validation_error(InvalidParamsError, kwargs):
                params = cls.parse_obj(kwargs)
            return await params.respond(managers)

        return JsonRpcMethod(respond, name=method)

    async def respond(self, managers: _ManagerWorkQueue) -> Any:
        raise NotImplementedError


class _ProfileParamMixin(BaseModel):
    profile: str


class _DefnParamMixin(BaseModel):
    defns: typing.List[Defn]


@_register_method('config/write')
class WriteConfigParams(BaseParams):
    values: typing.Dict[str, Any]
    infer_game_flavour: bool

    @t
    def respond(self, managers: _ManagerWorkQueue) -> Config:
        with _reraise_validation_error(_ConfigError):
            config = Config(**self.values)
            if self.infer_game_flavour:
                config.game_flavour = Config.infer_flavour(config.addon_dir)
            config.write()

        # Dispose of the ``Manager`` corresponding to the profile if any
        # so that the configuration is reloaded on next invocation
        managers.unload(config.profile)
        return config


@_register_method('config/read')
class ReadConfigParams(_ProfileParamMixin, BaseParams):
    @t
    def respond(self, managers: _ManagerWorkQueue) -> Config:
        with _reraise_validation_error(_ConfigError):
            return Config.read(self.profile)


@_register_method('config/delete')
class DeleteConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> None:
        async def delete_profile(manager: Manager):
            await t(manager.config.delete)()
            managers.unload(self.profile)

        await managers.run(self.profile, delete_profile)


@_register_method('config/list')
class ListProfilesParams(BaseParams):
    @t
    def respond(self, managers: _ManagerWorkQueue) -> list[str]:
        return Config.list_profiles()


class Source(TypedDict):
    source: str
    name: str
    supported_strategies: list[Strategy]
    supports_rollback: bool
    changelog_format: str


@_register_method('sources/list')
class ListSourcesParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> list[Source]:
        manager = await managers.run(self.profile, _get_manager)
        return [
            {
                'source': r.source,
                'name': r.name,
                'supported_strategies': sorted(r.strategies, key=list(Strategy).index),
                'supports_rollback': Strategy.version in r.strategies,
                'changelog_format': r.changelog_format.value,
            }
            for r in manager.resolvers.values()
        ]


@_register_method('list')
class ListInstalledParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> list[models.Pkg]:
        manager = await managers.run(self.profile, _get_manager)
        installed_pkgs = (
            manager.database.execute(sa.select(db.pkg).order_by(sa.func.lower(db.pkg.c.name)))
            .mappings()
            .all()
        )
        return [models.Pkg.from_row_mapping(manager.database, p) for p in installed_pkgs]


class SuccessResult(TypedDict):
    status: Literal['success']
    addon: models.Pkg


class ErrorResult(TypedDict):
    status: Literal['failure', 'error']
    message: str


@_register_method('search')
class SearchParams(_ProfileParamMixin, BaseParams):
    search_terms: str
    limit: int
    sources: typing.Set[str]
    start_date: typing.Optional[datetime]

    async def respond(self, managers: _ManagerWorkQueue) -> list[CatalogueEntry]:
        return await managers.run(
            self.profile,
            partial(
                Manager.search,
                search_terms=self.search_terms,
                limit=self.limit,
                sources=self.sources,
                start_date=self.start_date,
            ),
        )


@_register_method('resolve')
class ResolveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def _resolve(self, manager: Manager) -> dict[Defn, Any]:
        def extract_source(defn: Defn):
            if defn.source == '*':
                pair = manager.pair_uri(defn.alias)
                if pair:
                    source, alias = pair
                    defn = defn.with_(source=source, alias=alias)
            return defn

        return await manager.resolve(list(map(extract_source, self.defns)))

    async def respond(self, managers: _ManagerWorkQueue) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, self._resolve)
        return [
            {'status': 'success', 'addon': r}
            if models.is_pkg(r)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('install')
class InstallParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    replace: bool

    async def respond(self, managers: _ManagerWorkQueue) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.install, defns=self.defns, replace=self.replace)
        )
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('update')
class UpdateParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.update, defns=self.defns, retain_defn_strategy=True)
        )
        return [
            {'status': r.status, 'addon': r.new_pkg}
            if isinstance(r, R.PkgUpdated)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('remove')
class RemoveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    keep_folders: bool

    async def respond(self, managers: _ManagerWorkQueue) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.remove, defns=self.defns, keep_folders=self.keep_folders)
        )
        return [
            {'status': r.status, 'addon': r.old_pkg}
            if isinstance(r, R.PkgRemoved)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('pin')
class PinParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, partial(Manager.pin, defns=self.defns))
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('get_changelog')
class GetChangelogParams(_ProfileParamMixin, BaseParams):
    changelog_url: str

    async def respond(self, managers: _ManagerWorkQueue) -> str:
        return await managers.run(
            self.profile, partial(Manager.get_changelog, uri=self.changelog_url)
        )


class ReconcileResult(TypedDict):
    reconciled: list[ReconcileResult_AddonMatch]
    unreconciled: list[ReconcileResult_AddonMatch]


class ReconcileResult_AddonMatch(TypedDict):
    folders: list[ReconcileResult_AddonFolder]
    matches: list[models.Pkg]


class ReconcileResult_AddonFolder(TypedDict):
    name: str
    version: str


@_register_method('reconcile')
class ReconcileParams(_ProfileParamMixin, BaseParams):
    matcher: Literal['toc_source_ids', 'folder_name_subsets', 'addon_names_with_folder_names']

    async def respond(self, managers: _ManagerWorkQueue) -> ReconcileResult:
        leftovers = await managers.run(self.profile, t(matchers.get_unreconciled_folder_set))
        match_groups: matchers.FolderAndDefnPairs = await managers.run(
            self.profile, partial(getattr(matchers, f'match_{self.matcher}'), leftovers=leftovers)
        )
        uniq_defns = uniq(d for _, b in match_groups for d in b)
        resolve_results = await managers.run(
            self.profile, partial(Manager.resolve, defns=uniq_defns)
        )
        return {
            'reconciled': [
                {
                    'folders': [{'name': f.name, 'version': f.version} for f in a],
                    'matches': [r for i in d for r in (resolve_results[i],) if models.is_pkg(r)],
                }
                for a, d in match_groups
            ],
            'unreconciled': [
                {
                    'folders': [{'name': f.name, 'version': f.version}],
                    'matches': [],
                }
                for f in sorted(leftovers - frozenset(i for a, _ in match_groups for i in a))
            ],
        }


class DownloadProgressReport(TypedDict):
    defn: Defn
    progress: float


@_register_method('get_download_progress')
class GetDownloadProgressParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> list[DownloadProgressReport]:
        return [
            {'defn': Defn.from_pkg(p), 'progress': r}
            for p, r in await managers.get_download_progress(self.profile)
        ]


class GetVersionResult(TypedDict):
    installed_version: str
    new_version: str | None


@_register_method('meta/get_version')
class GetVersionParams(BaseParams):
    async def respond(self, managers: _ManagerWorkQueue) -> GetVersionResult:
        outdated, new_version = await is_outdated()
        return {
            'installed_version': __version__,
            'new_version': new_version if outdated else None,
        }


@_register_method('assist/open_url')
class OpenUrlParams(BaseParams):
    url: str

    async def respond(self, managers: _ManagerWorkQueue) -> None:
        click.launch(self.url)


@_register_method('assist/reveal_folder')
class RevealFolderParams(BaseParams):
    path_parts: typing.List[str]

    async def respond(self, managers: _ManagerWorkQueue) -> None:
        click.launch(os.path.join(*self.path_parts), locate=True)


class SelectFolderResult(TypedDict):
    selection: str | None


@_register_method('assist/select_folder')
class SelectFolderParams(BaseParams):
    initial_folder: typing.Optional[str]

    async def respond(self, managers: _ManagerWorkQueue) -> SelectFolderResult:
        from .app import InstawowApp

        try:
            (selection,) = InstawowApp.app.iw_window.select_folder_dialog(
                'Select folder', self.initial_folder
            )
        except ValueError:
            selection = None
        return {'selection': selection}


class ConfirmDialogueResult(TypedDict):
    ok: bool


@_register_method('assist/confirm')
class ConfirmDialogueParams(BaseParams):
    title: str
    message: str

    async def respond(self, managers: _ManagerWorkQueue) -> ConfirmDialogueResult:
        from .app import InstawowApp

        ok = InstawowApp.app.iw_window.confirm_dialog(self.title, self.message)
        return {'ok': ok}


def _init_json_rpc_web_client(
    progress_reporters: set[tuple[Manager, models.Pkg, Callable[[], float]]],
) -> aiohttp.ClientSession:
    async def do_on_request_end(
        client_session: aiohttp.ClientSession,
        trace_config_ctx: SimpleNamespace,
        params: aiohttp.TraceRequestEndParams,
    ) -> None:
        trace_request_ctx: TraceRequestCtx = trace_config_ctx.trace_request_ctx
        if (
            trace_request_ctx
            and trace_request_ctx['report_progress'] == 'pkg_download'
            and params.response.content_length
            and aiohttp.hdrs.CONTENT_ENCODING not in params.response.headers
        ):
            content = params.response.content
            content_length = params.response.content_length
            entry = (
                trace_request_ctx['manager'],
                trace_request_ctx['pkg'],
                lambda: content.total_bytes / content_length,  # type: ignore
            )
            progress_reporters.add(entry)
            content.on_eof(lambda: progress_reporters.remove(entry))

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return init_web_client(trace_configs=[trace_config])


async def _get_manager(manager: Manager):
    return manager


class _ManagerWorkQueue:
    def __init__(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._managers: dict[str, Manager] = {}
        self._queue: asyncio.Queue[_ManagerWorkQueueItem] = asyncio.Queue()
        self._progress_reporters: set[tuple[Manager, models.Pkg, Callable[[], float]]] = set()
        self._web_client = _init_json_rpc_web_client(self._progress_reporters)
        self._locks: defaultdict[str, asyncio.Lock] = defaultdict(asyncio.Lock)

    async def cleanup(self) -> None:
        for manager in self._managers.values():
            manager.database.close()
        await self._web_client.close()

    def unload(self, profile: str) -> None:
        manager = self._managers.pop(profile, None)
        if manager:
            manager.database.close()

    async def listen(self) -> None:
        async def schedule(
            future: asyncio.Future[Any], profile: str, coro_fn: Callable[..., Awaitable[Any]]
        ):
            try:
                async with self._locks[f"load profile '{profile}'"]:
                    try:
                        manager = self._managers[profile]
                    except KeyError:
                        with _reraise_validation_error(_ConfigError):
                            config = await t(Config.read)(profile)

                        manager = self._managers[profile] = Manager.from_config(config)

                result = await coro_fn(manager)
            except BaseException as exc:
                future.set_exception(exc)
            else:
                future.set_result(result)

        Manager.contextualise(web_client=self._web_client, locks=self._locks)

        while True:
            item = await self._queue.get()
            asyncio.create_task(schedule(*item))
            self._queue.task_done()

    async def run(
        self, profile: str, coro_fn: Callable[Concatenate[Manager, _P], Awaitable[_T]]
    ) -> _T:
        future = self._loop.create_future()
        self._queue.put_nowait((future, profile, coro_fn))
        return await asyncio.wait_for(future, None)

    async def get_download_progress(self, profile: str) -> Iterator[tuple[models.Pkg, float]]:
        manager = await self.run(profile, _get_manager)
        return ((p, r()) for m, p, r in self._progress_reporters if m is manager)


async def create_app() -> aiohttp.web.Application:
    managers = _ManagerWorkQueue()

    async def start_managers(app: aiohttp.web.Application):
        managers_listen = asyncio.create_task(managers.listen())
        yield
        managers_listen.cancel()
        await managers.cleanup()

    async def get_index(request: aiohttp.web.Request):
        return aiohttp.web.Response(
            content_type='text/html',
            text=importlib.resources.read_text(templates, 'index.html'),
        )

    async def get_static_file(request: aiohttp.web.Request):
        filename = request.path.lstrip('/')
        if filename.startswith('svelte-bundle.js'):
            content_type = 'application/javascript'
        elif filename == 'svelte-bundle.css':
            content_type = 'text/css'
        else:
            raise aiohttp.web.HTTPNotFound

        return aiohttp.web.Response(
            content_type=content_type,
            text=importlib.resources.read_text(frontend, filename),
        )

    def _json_serialize(value: dict[str, Any]) -> str:
        return BaseModel.construct(**value).json()

    rpc_server = WsJsonRpcServer(
        json_serialize=_json_serialize,
        middlewares=rpc_middlewares.DEFAULT_MIDDLEWARES,
    )
    rpc_server.add_methods([m.bind(n, managers) for n, m in methods])

    app = aiohttp.web.Application()
    app.add_routes(
        [
            aiohttp.web.get('/', get_index),
            aiohttp.web.get(
                r'/svelte-bundle{extension:(?:\.css|\.js(?:\.map)?)}', get_static_file
            ),
            aiohttp.web.get('/api', rpc_server.handle_http_request),
        ]
    )
    app.cleanup_ctx.append(start_managers)
    app.freeze()
    return app


async def prepare() -> tuple[URL, Callable[[], Awaitable[None]]]:
    "Fire up the server."
    app = await create_app()
    app_runner = aiohttp.web.AppRunner(app)
    await app_runner.setup()
    assert app_runner.server  # Server is created during setup
    # By omitting the port ``loop.create_server`` will find an available port
    # to bind to - equivalent to creating a socket on port 0.
    server = await asyncio.get_running_loop().create_server(app_runner.server, LOCALHOST)
    host, port = server.sockets[0].getsockname()

    async def serve():
        try:
            # ``server_forever`` cleans up after the server when it's interrupted
            await server.serve_forever()
        finally:
            await app_runner.cleanup()

    return (
        URL.build(scheme='http', host=host, port=port),
        serve,
    )



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from functools import partial, reduce
from itertools import chain, product
import time
import typing
from typing import TYPE_CHECKING, Any, TypeVar

from loguru import logger
from pydantic import BaseModel, Field, validator
from typing_extensions import Literal, Protocol, TypeAlias, TypedDict
from yarl import URL

from . import manager
from .config import BaseConfig, Flavour
from .utils import bucketise, chain_dict, gather
from .utils import run_in_thread as t
from .utils import shasum

# ``NotRequired`` is provisional and does not exist at runtime
if TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import NotRequired as N


_TAuras = TypeVar('_TAuras', bound='BaseAuras')
_Slug = str
_ImportString = str
_AuraGroup: TypeAlias = 'Sequence[tuple[Sequence[WeakAura], WagoApiResponse, _ImportString]]'

IMPORT_API_URL = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseConfig):
    wago_api_key: typing.Optional[str] = None


class BaseAuras(
    BaseModel,
    arbitrary_types_allowed=True,
    json_encoders={URL: str},
):
    class Meta(Protocol):
        api_url: URL
        filename: str

    @classmethod
    def from_lua_table(cls: type[_TAuras], lua_table: Any) -> _TAuras:
        raise NotImplementedError


def _merge_auras(auras: Iterable[_TAuras]) -> dict[type, _TAuras]:
    "Merge auras of the same type."
    return {
        t: t(__root__=reduce(lambda a, b: {**a, **b}, (i.__root__ for i in a)))
        for t, a in bucketise(auras, key=type).items()
    }


class WeakAura(
    BaseModel,
    allow_population_by_field_name=True,
    arbitrary_types_allowed=True,
):
    id: str
    uid: str
    parent: typing.Optional[str]
    url: URL
    version: int

    @classmethod
    def from_lua_table(cls, lua_table: Any) -> WeakAura | None:
        url_string = lua_table.get('url')
        if url_string is not None:
            url = URL(url_string)
            if url.host == 'wago.io':
                return cls.parse_obj({**lua_table, 'url': url})

    @validator('url', pre=True)
    def _url_to_URL(cls, value: str | URL) -> URL:
        if not isinstance(value, URL):
            value = URL(value)
        return value


class WeakAuras(BaseAuras):
    class Meta:
        api_url = URL('https://data.wago.io/api/check/weakauras')
        filename = 'WeakAuras.lua'

    __root__: typing.Dict[_Slug, typing.List[WeakAura]]

    @classmethod
    def from_lua_table(cls, lua_table: Any) -> WeakAuras:
        auras = (
            a for t in lua_table['displays'].values() for a in (WeakAura.from_lua_table(t),) if a
        )
        sorted_auras = sorted(auras, key=lambda a: a.id)
        return cls(__root__=bucketise(sorted_auras, key=lambda a: a.url.parts[1]))


class Plateroo(WeakAura):
    id: str = Field(alias='Name')
    uid = ''


class Plateroos(BaseAuras):
    class Meta:
        api_url = URL('https://data.wago.io/api/check/plater')
        filename = 'Plater.lua'

    __root__: typing.Dict[_Slug, typing.List[Plateroo]]

    @classmethod
    def from_lua_table(cls, lua_table: Any) -> Plateroos:
        auras = (
            a
            for n, p in lua_table['profiles'].items()
            for t in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                p.get('script_data') or (),
                p.get('hook_data') or (),
            )
            for a in (Plateroo.from_lua_table(t),)
            if a
        )
        sorted_auras = sorted(auras, key=lambda a: a.id)
        return cls(__root__={a.url.parts[1]: [a] for a in sorted_auras})


class WagoApiResponse(TypedDict):
    _id: str  # +   # Alphanumeric ID
    name: str  # +  # User-facing name
    slug: str  # +  # Slug if it has one; otherwise same as ``_id``
    url: str
    created: str  # ISO datetime
    modified: str  # ISO datetime
    game: str  # "classic" or xpac, e.g. "bfa"
    username: N[str]  # +  # Author username
    version: int  # +   # Version counter, incremented with every update
    # Semver auto-generated from ``version`` - for presentation only
    versionString: str
    changelog: WagoApiResponse_Changelog  # +
    forkOf: N[str]  # Only present on forks
    regionType: N[str]  # Only present on WAs


class WagoApiResponse_Changelog(TypedDict):
    format: N[Literal['bbcode', 'markdown']]
    text: N[str]


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: manager.Manager, builder_config: BuilderConfig) -> None:
        self.manager = manager
        self.builder_config = builder_config

        output_folder = self.manager.config.plugin_dir / __name__
        self.addon_zip_path = output_folder / 'WeakAurasCompanion.zip'
        self.changelog_path = output_folder / 'CHANGELOG.md'
        self.checksum_txt_path = output_folder / 'checksum.txt'

    @staticmethod
    def extract_auras(model: type[_TAuras], source: str) -> _TAuras:
        from ._custom_slpp import SLPP

        source_after_assignment = source[source.find('=') + 1 :]
        lua_table = SLPP(source_after_assignment).decode()
        return model.from_lua_table(lua_table)

    def extract_installed_auras(self) -> Iterator[WeakAuras | Plateroos]:
        flavour_root = self.manager.config.addon_dir.parents[1]
        saved_vars_of_every_account = flavour_root.glob('WTF/Account/*/SavedVariables')
        for saved_vars, model in product(
            saved_vars_of_every_account,
            [WeakAuras, Plateroos],
        ):
            file = saved_vars / model.Meta.filename
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                content = file.read_text(encoding='utf-8-sig', errors='replace')
                aura_group_cache = self.manager.config.cache_dir / shasum(content)
                if aura_group_cache.exists():
                    logger.info(f'loading {file} from cache at {aura_group_cache}')
                    aura_groups = model.parse_file(aura_group_cache)
                else:
                    start = time.perf_counter()
                    aura_groups = self.extract_auras(model, content)
                    logger.debug(f'{model.__name__} extracted in {time.perf_counter() - start}s')
                    aura_group_cache.write_text(aura_groups.json(), encoding='utf-8')
                yield aura_groups

    async def _fetch_wago_metadata(
        self, api_url: URL, aura_ids: Iterable[str]
    ) -> list[WagoApiResponse]:
        from aiohttp import ClientResponseError

        try:
            return sorted(
                await manager.cache_response(
                    self.manager,
                    api_url.with_query(ids=','.join(aura_ids)),
                    {'minutes': 30},
                    label='Fetching aura metadata',
                    request_extra={'headers': {'api-key': self.builder_config.wago_api_key or ''}},
                ),
                key=lambda r: r['slug'],
            )
        except ClientResponseError as error:
            if error.status != 404:
                raise
            return []

    async def _fetch_wago_import_string(self, aura: WagoApiResponse) -> str:
        return await manager.cache_response(
            self.manager,
            IMPORT_API_URL.with_query(id=aura['_id']).with_fragment(str(aura['version'])),
            {'days': 30},
            label=f"Fetching aura '{aura['slug']}'",
            is_json=False,
            request_extra={'headers': {'api-key': self.builder_config.wago_api_key or ''}},
        )

    async def get_remote_auras(self, auras: BaseAuras) -> _AuraGroup:
        if not auras.__root__:
            return []

        metadata = await self._fetch_wago_metadata(auras.Meta.api_url, auras.__root__)
        import_strings = await gather(self._fetch_wago_import_string(r) for r in metadata)
        return [(auras.__root__[r['slug']], r, i) for r, i in zip(metadata, import_strings)]

    def _checksum(self) -> str:
        from hashlib import sha256

        return sha256(self.addon_zip_path.read_bytes()).hexdigest()

    async def get_checksum(self) -> str:
        return await t(self.checksum_txt_path.read_text)(encoding='utf-8')

    def _get_toc_number(self) -> str:
        game_flavour: Flavour = self.manager.config.game_flavour
        if game_flavour is Flavour.retail:
            return '90100'
        elif game_flavour is Flavour.vanilla_classic:
            return '11400'
        elif game_flavour is Flavour.burning_crusade_classic:
            return '20502'

    def _generate_addon(
        self, auras: Iterable[tuple[type[WeakAuras | Plateroos], _AuraGroup]]
    ) -> None:
        from importlib.resources import read_text
        from zipfile import ZipFile, ZipInfo

        from jinja2 import Environment, FunctionLoader

        from . import wa_templates

        jinja_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FunctionLoader(partial(read_text, wa_templates)),
        )
        aura_dict = chain_dict((WeakAuras, Plateroos), (), auras)

        self.addon_zip_path.parent.mkdir(exist_ok=True)
        with ZipFile(self.addon_zip_path, 'w') as file:

            def write_tpl(filename: str, ctx: dict[str, Any]):
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp would be set to the current time
                # which would render the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                file.writestr(zip_info, jinja_env.get_template(filename).render(ctx))

            write_tpl(
                'data.lua',
                {
                    'weakauras': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata.get('username', '__unknown__'),
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                # ``wagoSemver`` is supposed to be the ``versionString``
                                # from Wago but there is a bug where the ``version``
                                # is sometimes not appended to the semver.
                                # The Companion add-on's version is derived from its checksum
                                # so if ``wagoSemver`` were to change between requests
                                # we'd be triggering spurious updates in instawow.
                                'wagoSemver': metadata['version'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[WeakAuras]
                    ],
                    # Maps internal UIDs of top-level auras to IDs or slugs on Wago
                    'weakaura_uids': [
                        (a.uid, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                    ],
                    # Maps local names to IDs or slugs on Wago
                    'weakaura_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in existing_auras
                    ],
                    'plateroos': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata.get('username', '__unknown__'),
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                'wagoSemver': metadata['version'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[Plateroos]
                    ],
                    'plater_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[Plateroos]
                        for a in existing_auras
                    ],
                },
            )
            write_tpl('init.lua', {})
            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': self._get_toc_number()},
            )

        self.changelog_path.write_text(
            jinja_env.get_template(self.changelog_path.name).render(
                {
                    'changelog_entries': [
                        (
                            a.id,
                            a.url.parent,
                            metadata['version'],
                            metadata['changelog'].get('text') or 'n/a',
                        )
                        for v in aura_dict.values()
                        for existing_auras, metadata, _ in v
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                        if a.version != metadata['version']
                    ]
                }
            )
            or 'n/a',
            encoding='utf-8',
        )

        self.checksum_txt_path.write_text(
            self._checksum(),
            encoding='utf-8',
        )

    async def build(self) -> None:
        installed_auras = await t(list)(self.extract_installed_auras())
        installed_auras_by_type = _merge_auras(installed_auras)
        aura_groups = await gather(map(self.get_remote_auras, installed_auras_by_type.values()))
        await t(self._generate_addon)(zip(installed_auras_by_type, aura_groups))



from __future__ import annotations

from collections.abc import AsyncIterator, Sequence, Set
from datetime import datetime, timezone
from enum import IntEnum
from itertools import chain, takewhile
import re
import typing
from typing import TYPE_CHECKING, Any, ClassVar

from loguru import logger
from pydantic import BaseModel
from typing_extensions import Literal, Protocol, TypedDict
from yarl import URL

from . import _deferred_types, manager, models
from . import results as R
from .common import Strategy
from .config import Flavour
from .utils import StrEnum, bucketise, cached_property, gather, normalise_names, uniq

if TYPE_CHECKING:  # pragma: no cover
    from typing_extensions import NotRequired as N


class ChangelogFormat(StrEnum):
    html = 'html'
    markdown = 'markdown'
    bbcode = 'bbcode'
    raw = 'raw'


class Defn(
    BaseModel,
    frozen=True,
):
    source: str
    alias: str
    id: typing.Optional[str] = None
    strategy: Strategy = Strategy.default
    version: typing.Optional[str] = None

    def __init__(self, source: str, alias: str, **kwargs: Any) -> None:
        super().__init__(source=source, alias=alias, **kwargs)

    @classmethod
    def from_pkg(cls, pkg: models.Pkg) -> Defn:
        return cls(
            source=pkg.source,
            alias=pkg.slug,
            id=pkg.id,
            strategy=pkg.options.strategy,
            version=pkg.version,
        )

    def with_(self, **kwargs: Any) -> Defn:
        return self.__class__(**{**self.__dict__, **kwargs})

    def with_version(self, version: str) -> Defn:
        return self.with_(strategy=Strategy.version, version=version)

    def to_urn(self) -> str:
        return f'{self.source}:{self.alias}'


slugify = normalise_names('-')


def _format_data_changelog(changelog: str = '') -> str:
    import urllib.parse

    return f'data:,{urllib.parse.quote(changelog)}'


class CatatalogueBaseEntry(BaseModel):
    source: str
    id: str
    slug: str = ''
    name: str
    game_flavours: typing.Set[Flavour]
    folders: typing.List[typing.Set[str]] = []
    download_count: int
    last_updated: datetime


class CatalogueEntry(CatatalogueBaseEntry):
    derived_download_score: float


class Catalogue(
    BaseModel,
    json_encoders={set: sorted},
    keep_untouched=(cached_property,),
):
    __root__: typing.List[CatalogueEntry]

    @classmethod
    async def collate(cls, start_date: datetime | None) -> Catalogue:
        async with manager.init_web_client() as web_client:
            items = [a for r in manager.Manager.RESOLVERS async for a in r.catalogue(web_client)]

        most_downloads_per_source = {
            s: max(e.download_count for e in i)
            for s, i in bucketise(items, key=lambda v: v.source).items()
        }
        entries = (
            CatalogueEntry(
                **i.__dict__,
                derived_download_score=i.download_count / most_downloads_per_source[i.source],
            )
            for i in items
        )
        if start_date is not None:
            entries = (e for e in entries if e.last_updated >= start_date)
        catalogue = cls.parse_obj(list(entries))
        return catalogue

    @cached_property
    def curse_slugs(self) -> dict[str, str]:
        return {a.slug: a.id for a in self.__root__ if a.source == 'curse'}


class Resolver(Protocol):
    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategy]]
    changelog_format: ClassVar[ChangelogFormat]

    def __init__(self, manager: manager.Manager) -> None:
        ...

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        "Attempt to extract a ``Defn`` alias from a given URL."

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, models.Pkg | R.ManagerError | R.InternalError]:
        "Resolve multiple ``Defn``s into packages."

    async def resolve_one(self, defn: Defn, metadata: Any) -> models.Pkg:
        "Resolve a ``Defn`` into a package."

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        "Yield add-ons from source for cataloguing."
        return
        yield


class BaseResolver(Resolver):
    def __init__(self, manager: manager.Manager) -> None:
        self.manager = manager

    def __init_subclass__(cls) -> None:
        orig_resolve_one = cls.resolve_one

        async def resolve_one(self: BaseResolver, defn: Defn, metadata: Any) -> models.Pkg:
            if defn.strategy not in self.strategies:
                raise R.PkgStrategyUnsupported(defn.strategy)

            return await orig_resolve_one(self, defn, metadata)

        cls.resolve_one = resolve_one  # type: ignore

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        return None

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, models.Pkg | R.ManagerError | R.InternalError]:
        results = await gather(
            (self.resolve_one(d, None) for d in defns), manager.capture_manager_exc_async
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: Any) -> models.Pkg:
        raise NotImplementedError

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        return
        yield


_CurseFlavor = Literal['wow_burning_crusade', 'wow_classic', 'wow_retail']


class _CurseGameVersionTypeId(IntEnum):
    retail = 517
    vanilla_classic = 67408
    burning_crusade_classic = 73246


# Only documenting the fields we're actually using.
class _CurseAddon(TypedDict):
    id: int
    name: str  # User-facing add-on name
    websiteUrl: str  # e.g. 'https://www.curseforge.com/wow/addons/molinari'
    summary: str  # One-line description of the add-on
    downloadCount: int  # Total number of downloads
    latestFiles: list[_CurseAddon_File]
    gameVersionLatestFiles: list[_CurseAddon_GameVersionFile]
    slug: str  # URL slug; 'molinari' in 'https://www.curseforge.com/wow/addons/molinari'
    dateReleased: str  # ISO datetime of latest release


class _CurseAddon_File(TypedDict):
    id: int  # Unique file ID
    displayName: str  # Tends to be the version
    downloadUrl: str
    fileDate: str  # Upload datetime in ISO, e.g. '2020-02-02T12:12:12Z'
    releaseType: Literal[1, 2, 3]  # 1 = stable; 2 = beta; 3 = alpha
    dependencies: list[_CurseAddon_FileDependency]
    modules: list[_CurseAddon_FileModules]
    exposeAsAlternative: bool | None
    gameVersionFlavor: _CurseFlavor
    sortableGameVersion: list[_CurseAddon_FileSortableGameVersion]


class _CurseAddon_GameVersionFile(TypedDict):
    projectFileName: str
    gameVersionFlavor: _CurseFlavor
    gameVersionTypeId: N[_CurseGameVersionTypeId]


class _CurseAddon_FileDependency(TypedDict):
    id: int  # Unique dependency ID
    addonId: int  # The ID of the add-on we're depending on
    # The type of dependency.  One of:
    #   1 = embedded library
    #   2 = optional dependency
    #   3 = required dependency (this is the one we're after)
    #   4 = tool
    #   5 = incompatible
    #   6 = include (wat)
    type: Literal[1, 2, 3, 4, 5, 6]
    fileId: int  # The ID of the parent file which has this as a dependency


class _CurseAddon_FileModules(TypedDict):
    foldername: str
    fingerprint: int  # Folder fingerprint used by Curse for reconciliation
    # One of:
    #   1 = package
    #   2 = module
    #   3 = main module
    #   4 = file
    #   5 = referenced file
    # For WoW add-ons the main folder will have type "3" and the rest
    # of them type "2"
    type: Literal[1, 2, 3, 4, 5]


class _CurseAddon_FileSortableGameVersion(TypedDict):
    gameVersionTypeId: N[_CurseGameVersionTypeId]


class _CurseFile(TypedDict):
    id: int
    displayName: str
    downloadUrl: str
    fileDate: str  # Upload datetime in ISO, e.g. '2020-02-02T12:12:12Z'
    releaseType: Literal[1, 2, 3]  # 1 = stable; 2 = beta; 3 = alpha
    dependencies: list[_CurseAddon_FileDependency]
    modules: list[_CurseAddon_FileModules]
    gameVersionFlavor: Literal['wow_burning_crusade', 'wow_classic', 'wow_retail']


class CurseResolver(BaseResolver):
    source = 'curse'
    name = 'CurseForge'
    strategies = frozenset(
        {
            Strategy.default,
            Strategy.latest,
            Strategy.any_flavour,
            Strategy.version,
        }
    )
    changelog_format = ChangelogFormat.html

    # Reference: https://twitchappapi.docs.apiary.io/
    addon_api_url = URL('https://addons-ecs.forgesvc.net/api/v2/addon')

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        if url.host == 'www.wowace.com' and len(url.parts) > 2 and url.parts[1] == 'projects':
            return url.parts[2].lower()
        elif (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, models.Pkg | R.ManagerError | R.InternalError]:
        from aiohttp import ClientResponseError

        catalogue = await self.manager.synchronise()

        defns_to_ids = {d: d.id or catalogue.curse_slugs.get(d.alias) or d.alias for d in defns}
        numeric_ids = uniq(i for i in defns_to_ids.values() if i.isdigit())
        try:
            json_response: list[_CurseAddon] = await manager.cache_response(
                self.manager,
                self.addon_api_url,
                {'minutes': 5},
                request_extra={'method': 'POST', 'json': numeric_ids},
            )
        except ClientResponseError as error:
            if error.status != 404:
                raise
            json_response = []

        api_results = {str(r['id']): r for r in json_response}
        results = await gather(
            (self.resolve_one(d, api_results.get(i)) for d, i in defns_to_ids.items()),
            manager.capture_manager_exc_async,
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _CurseAddon | None) -> models.Pkg:
        if metadata is None:
            raise R.PkgNonexistent

        if defn.strategy is Strategy.version:
            all_files: list[_CurseFile] = await manager.cache_response(
                self.manager,
                self.addon_api_url / str(metadata['id']) / 'files',
                {'hours': 1},
                label=f'Fetching metadata from {self.name}',
            )

            file = next((f for f in all_files if defn.version == f['displayName']), None)
            if file is None:
                raise R.PkgFileUnavailable(f'version {defn.version} not found')

        else:
            latest_files = metadata['latestFiles']
            if not latest_files:
                raise R.PkgFileUnavailable('no files available for download')

            def make_filter():
                def is_not_libless(f: _CurseAddon_File):
                    # There's also an 'isAlternate' field that's missing from some
                    # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
                    # absent from the /file endpoint
                    return not f['exposeAsAlternative']

                yield is_not_libless

                if defn.strategy is not Strategy.any_flavour:

                    def supports_x(
                        curse_flavor: _CurseFlavor, curse_type_id: _CurseGameVersionTypeId
                    ):
                        def excise_flavour(f: _CurseAddon_File):
                            return f['gameVersionFlavor'] == curse_flavor or any(
                                s.get('gameVersionTypeId') == curse_type_id
                                for s in f['sortableGameVersion']
                            )

                        return excise_flavour

                    if self.manager.config.game_flavour is Flavour.retail:
                        yield supports_x(
                            'wow_retail',
                            _CurseGameVersionTypeId.retail,
                        )
                    elif self.manager.config.game_flavour is Flavour.vanilla_classic:
                        yield supports_x(
                            'wow_classic',
                            _CurseGameVersionTypeId.vanilla_classic,
                        )
                    elif self.manager.config.game_flavour is Flavour.burning_crusade_classic:
                        yield supports_x(
                            'wow_burning_crusade',
                            _CurseGameVersionTypeId.burning_crusade_classic,
                        )

                if defn.strategy is not Strategy.latest:

                    def is_stable_release(f: _CurseAddon_File):
                        return f['releaseType'] == 1

                    yield is_stable_release

            filter_fns = list(make_filter())

            file = max(
                (f for f in latest_files if all(l(f) for l in filter_fns)),
                # The ``id`` is just a counter so we don't have to go digging around dates
                key=lambda f: f['id'],
                default=None,
            )
            if file is None:
                raise R.PkgFileUnavailable(
                    f'no files matching {self.manager.config.game_flavour} '
                    f'using {defn.strategy} strategy'
                )

        return models.Pkg(
            source=self.source,
            id=metadata['id'],
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['websiteUrl'],
            download_url=file['downloadUrl'],
            date_published=file['fileDate'],
            version=file['displayName'],
            changelog_url=str(
                self.addon_api_url / str(metadata['id']) / 'file' / str(file['id']) / 'changelog'
            ),
            options={'strategy': defn.strategy},
            deps=[{'id': d['addonId']} for d in file['dependencies'] if d['type'] == 3],
        )

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        def supports_x(files: list[_CurseAddon_File]):
            def excise_flavour(
                curse_flavor: _CurseFlavor,
                curse_type_id: _CurseGameVersionTypeId,
                flavour: Flavour,
            ):
                if any(
                    f['gameVersionFlavor'] == curse_flavor
                    or any(
                        s.get('gameVersionTypeId') == curse_type_id
                        for s in f['sortableGameVersion']
                    )
                    for f in files
                ):
                    yield flavour

            yield from excise_flavour(
                'wow_retail',
                _CurseGameVersionTypeId.retail,
                Flavour.retail,
            )
            yield from excise_flavour(
                'wow_classic',
                _CurseGameVersionTypeId.vanilla_classic,
                Flavour.vanilla_classic,
            )
            yield from excise_flavour(
                'wow_burning_crusade',
                _CurseGameVersionTypeId.burning_crusade_classic,
                Flavour.burning_crusade_classic,
            )

        step = 50
        sort_order = '3'  # Alphabetical
        for index in range(0, 10001 - step, step):
            async with web_client.get(
                (cls.addon_api_url / 'search').with_query(
                    gameId='1', sort=sort_order, pageSize=step, index=index
                )
            ) as response:
                items: list[_CurseAddon] = await response.json()

            if not items:
                break

            for item in items:
                folders = uniq(
                    frozenset(m['foldername'] for m in f['modules'])
                    for f in item['latestFiles']
                    if not f['exposeAsAlternative']
                )
                yield CatatalogueBaseEntry(
                    source=cls.source,
                    id=item['id'],
                    slug=item['slug'],
                    name=item['name'],
                    game_flavours=supports_x(item['latestFiles']),
                    folders=folders,
                    download_count=item['downloadCount'],
                    last_updated=item['dateReleased'],
                )


class _WowiCommonTerms(TypedDict):
    UID: str  # Unique add-on ID
    UICATID: str  # ID of category add-on is placed in
    UIVersion: str  # Add-on version
    UIDate: int  # Upload date expressed as unix epoch
    UIName: str  # User-facing add-on name
    UIAuthorName: str


class _WowiListApiItem_CompatibilityEntry(TypedDict):
    version: str  # Game version, e.g. '8.3.0'
    name: str  # Xpac or patch name, e.g. "Visions of N'Zoth" for 8.3.0


class _WowiListApiItem(_WowiCommonTerms):
    UIFileInfoURL: str  # Add-on page on WoWI
    UIDownloadTotal: str  # Total number of downloads
    UIDownloadMonthly: str  # Number of downloads in the last month and not 'monthly'
    UIFavoriteTotal: str
    UICompatibility: list[_WowiListApiItem_CompatibilityEntry] | None  # ``null`` if would be empty
    UIDir: list[str]  # Names of folders contained in archive
    UIIMG_Thumbs: list[str] | None  # Thumbnail URLs; ``null`` if would be empty
    UIIMGs: list[str] | None  # Full-size image URLs; ``null`` if would be empty
    # There are only two add-ons on the entire list with siblings
    # (they refer to each other). I don't know if this was meant to capture
    # dependencies (probably not) but it's so underused as to be worthless.
    # ``null`` if would be empty
    UISiblings: list[str] | None
    UIDonationLink: str | None  # Absent from the first item on the list (!)


class _WowiDetailsApiItem(_WowiCommonTerms):
    UIMD5: str | None  # Archive hash, ``null` when UI is pending
    UIFileName: str  # The actual filename, e.g. 'foo.zip'
    UIDownload: str  # Download URL
    UIPending: Literal['0', '1']  # Set to '1' if the file is awaiting approval
    UIDescription: str  # Long description with BB Code and all
    UIChangeLog: str  # This can also contain BB Code
    UIHitCount: str  # Same as UIDownloadTotal
    UIHitCountMonthly: str  # Same as UIDownloadMonthly


class _WowiCombinedItem(_WowiListApiItem, _WowiDetailsApiItem):
    pass


class WowiResolver(BaseResolver):
    source = 'wowi'
    name = 'WoWInterface'
    strategies = frozenset({Strategy.default})
    changelog_format = ChangelogFormat.bbcode

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the as yet unreleased Minion v4,
    # which is fair to assume is unstable.  They changed the naming scheme to
    # camelCase and some fields which were strings were converted to numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # 'UICompatibility' can't be relied on to enforce compatibility
    # in instawow.  The API appears to inherit the version of the latest
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever the download link always points to the
    # 'retail' version, which for single-file add-ons belonging to the
    # classic category would be an add-on for classic.
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                return url.query.get('fileid')
            elif url.name == 'fileinfo.php':
                return url.query.get('id')
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                return match and match['id']

    async def _synchronise(self) -> dict[str, _WowiListApiItem]:
        async with self.manager.locks['load WoWI catalogue']:
            list_api_items = await manager.cache_response(
                self.manager,
                self.list_api_url,
                {'hours': 1},
                label=f'Synchronising {self.name} catalogue',
            )
            return {i['UID']: i for i in list_api_items}

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, models.Pkg | R.ManagerError | R.InternalError]:
        from aiohttp import ClientResponseError

        list_api_items = await self._synchronise()

        defns_to_ids = {d: ''.join(takewhile(str.isdigit, d.alias)) for d in defns}
        numeric_ids = frozenset(filter(None, defns_to_ids.values()))
        try:
            details_api_items: list[_WowiDetailsApiItem] = await manager.cache_response(
                self.manager,
                self.details_api_url / f'{",".join(numeric_ids)}.json',
                {'minutes': 5},
            )
        except ClientResponseError as error:
            if error.status != 404:
                raise
            details_api_items = []

        combined_items: dict[str, Any] = {
            i['UID']: {**list_api_items[i['UID']], **i} for i in details_api_items
        }
        results = await gather(
            (self.resolve_one(d, combined_items.get(i)) for d, i in defns_to_ids.items()),
            manager.capture_manager_exc_async,
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _WowiCombinedItem | None) -> models.Pkg:
        if metadata is None:
            raise R.PkgNonexistent

        return models.Pkg(
            source=self.source,
            id=metadata['UID'],
            slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=metadata['UIFileInfoURL'],
            download_url=metadata['UIDownload'],
            date_published=metadata['UIDate'],
            version=metadata['UIVersion'],
            changelog_url=_format_data_changelog(metadata['UIChangeLog']),
            options={'strategy': defn.strategy},
        )

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        flavours = set(Flavour)

        async with web_client.get(cls.list_api_url) as response:
            items: list[_WowiListApiItem] = await response.json()

        for item in items:
            yield CatatalogueBaseEntry(
                source=cls.source,
                id=item['UID'],
                name=item['UIName'],
                folders=[item['UIDir']],
                game_flavours=flavours,
                download_count=item['UIDownloadTotal'],
                last_updated=item['UIDate'],
            )


class _TukuiUi(TypedDict):
    author: str
    category: str
    changelog: str
    donate_url: str
    downloads: int
    git: str
    id: Literal[-1, -2]  # -1 is Tukui and -2 ElvUI
    lastdownload: str
    lastupdate: str  # ISO date and no tz, e.g. '2020-02-02'
    name: str
    patch: str
    screenshot_url: str
    small_desc: str
    ticket: str
    url: str
    version: str
    web_url: str


class _TukuiAddon(TypedDict):
    author: str
    category: str
    changelog: str
    donate_url: str
    downloads: str  # Not a mistake, it is actually a string
    id: str
    last_download: str
    # ISO *datetime* with space sep and without an offset, e.g. '2020-02-02 12:12:20'
    lastupdate: str
    name: str
    patch: str | None
    screenshot_url: str
    small_desc: str
    url: str
    version: str
    web_url: str


class TukuiResolver(BaseResolver):
    source = 'tukui'
    name = 'Tukui'
    strategies = frozenset({Strategy.default})
    changelog_format = ChangelogFormat.html

    # There's also a ``/client-api.php`` endpoint which is apparently
    # used by the Tukui client itself to check for updates for the two retail
    # UIs only.  The response body appears to be identical to ``/api.php``
    api_url = URL('https://www.tukui.org/api.php')

    retail_ui_suites = {'elvui', 'tukui'}

    query_flavours = {
        Flavour.retail: 'addons',
        Flavour.vanilla_classic: 'classic-addons',
        Flavour.burning_crusade_classic: 'classic-tbc-addons',
    }

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        if url.host == 'www.tukui.org':
            if url.path in {'/addons.php', '/classic-addons.php', '/classic-tbc-addons.php'}:
                return url.query.get('id')
            elif url.path == '/download.php':
                return url.query.get('ui')

    async def _synchronise(self) -> dict[str, _TukuiAddon | _TukuiUi]:
        async def fetch_ui(ui_slug: str):
            addon: _TukuiUi = await manager.cache_response(
                self.manager,
                self.api_url.with_query({'ui': ui_slug}),
                {'minutes': 5},
            )
            return [(str(addon['id']), addon), (ui_slug, addon)]

        async def fetch_addons(flavour: Flavour):
            addons: list[_TukuiAddon] = await manager.cache_response(
                self.manager,
                self.api_url.with_query({self.query_flavours[flavour]: 'all'}),
                {'minutes': 30},
                label=f'Synchronising {self.name} {flavour} catalogue',
            )
            return [(str(a['id']), a) for a in addons]

        async with self.manager.locks['load Tukui catalogue']:
            addon_lists = await gather(
                [
                    *map(
                        fetch_ui,
                        self.retail_ui_suites
                        if self.manager.config.game_flavour is Flavour.retail
                        else [],
                    ),
                    fetch_addons(self.manager.config.game_flavour),
                ]
            )
            return {k: v for l in addon_lists for k, v in l}

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, models.Pkg | R.ManagerError | R.InternalError]:
        addons = await self._synchronise()
        ids = (
            d.alias[:p] if d.alias not in self.retail_ui_suites and p != -1 else d.alias
            for d in defns
            for p in (d.alias.find('-', 1),)
        )
        results = await gather(
            (self.resolve_one(d, addons.get(i)) for d, i in zip(defns, ids)),
            manager.capture_manager_exc_async,
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: _TukuiAddon | _TukuiUi | None) -> models.Pkg:
        if metadata is None:
            raise R.PkgNonexistent

        if metadata['id'] == -1:
            slug = 'tukui'
        elif metadata['id'] == -2:
            slug = 'elvui'
        else:
            slug = slugify(f'{metadata["id"]} {metadata["name"]}')

        return models.Pkg(
            source=self.source,
            id=str(metadata['id']),
            slug=slug,
            name=metadata['name'],
            description=metadata['small_desc'],
            url=metadata['web_url'],
            download_url=metadata['url'],
            date_published=datetime.fromisoformat(metadata['lastupdate']).replace(
                tzinfo=timezone.utc
            ),
            version=metadata['version'],
            changelog_url=(
                # The changelog URL is not versioned - adding fragment to allow caching
                str(URL(metadata['changelog']).with_fragment(metadata['version']))
                if metadata['id'] in {-1, -2}
                # Regular add-ons don't have dedicated changelogs but rather
                # link to the changelog tab on the add-on page
                else _format_data_changelog(metadata['changelog'])
            ),
            options={'strategy': defn.strategy},
        )

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        async def fetch_ui(ui_slug: str) -> list[_TukuiUi]:
            async with web_client.get(cls.api_url.with_query({'ui': ui_slug})) as response:
                return [await response.json(content_type=None)]  # text/html

        async def fetch_addons(query: str) -> list[_TukuiAddon]:
            async with web_client.get(cls.api_url.with_query({query: 'all'})) as response:
                return await response.json(content_type=None)  # text/html

        for flavours, item_coro in [
            ({Flavour.retail}, fetch_ui('tukui')),
            ({Flavour.retail}, fetch_ui('elvui')),
            ({Flavour.retail}, fetch_addons('addons')),
            ({Flavour.vanilla_classic}, fetch_addons('classic-addons')),
            ({Flavour.burning_crusade_classic}, fetch_addons('classic-tbc-addons')),
        ]:
            for item in await item_coro:
                yield CatatalogueBaseEntry(
                    source=cls.source,
                    id=item['id'],
                    name=item['name'],
                    game_flavours=flavours,
                    # Split Tukui and ElvUI downloads evenly between them.
                    # They both have the exact same number of downloads so
                    # I'm assuming they're being counted together.
                    # This should help with scoring other add-ons on the
                    # Tukui catalogue higher
                    download_count=int(item['downloads']) // (2 if item['id'] in {-1, -2} else 1),
                    last_updated=datetime.fromisoformat(item['lastupdate']).replace(
                        tzinfo=timezone.utc
                    ),
                )


# Not exhaustive (as you might've guessed).  Reference:
# https://docs.github.com/en/rest/reference/repos
class _GithubRepo(TypedDict):
    name: str  # the repo in user-or-org/repo
    full_name: str  # user-or-org/repo
    description: str
    html_url: str


class _GithubRelease(TypedDict):
    tag_name: str  # Hopefully the version
    published_at: str  # ISO datetime
    assets: list[_GithubRelease_Asset]
    body: str


class _GithubRelease_Asset(TypedDict):
    name: str  # filename
    content_type: str  # mime type
    state: Literal['starter', 'uploaded']
    browser_download_url: str


class _PackagerReleaseJson(TypedDict):
    releases: list[_PackagerReleaseJson_Release]


class _PackagerReleaseJson_Release(TypedDict):
    filename: str
    nolib: bool
    metadata: list[_PackagerReleaseJson_Release_Metadata]


class _PackagerReleaseJson_Release_Metadata(TypedDict):
    flavor: Literal['mainline', 'classic', 'bcc']
    interface: int


class GithubResolver(BaseResolver):
    source = 'github'
    name = 'GitHub'
    strategies = frozenset({Strategy.default, Strategy.latest, Strategy.version})
    changelog_format = ChangelogFormat.markdown

    repos_api_url = URL('https://api.github.com/repos')

    release_json_flavours = {
        Flavour.retail: 'mainline',
        Flavour.vanilla_classic: 'classic',
        Flavour.burning_crusade_classic: 'bcc',
    }

    @staticmethod
    def get_alias_from_url(url: URL) -> str | None:
        if url.host == 'github.com' and len(url.parts) > 2:
            return '/'.join(url.parts[1:3])

    async def resolve_one(self, defn: Defn, metadata: None) -> models.Pkg:
        from aiohttp import ClientResponseError

        repo_url = self.repos_api_url / defn.alias

        try:
            project_metadata: _GithubRepo = await manager.cache_response(
                self.manager, repo_url, {'hours': 1}
            )
        except ClientResponseError as error:
            if error.status == 404:
                raise R.PkgNonexistent
            raise

        if defn.strategy is Strategy.version:
            assert defn.version
            release_url = repo_url / 'releases/tags' / defn.version
        elif defn.strategy is Strategy.latest:
            release_url = (repo_url / 'releases').with_query(per_page='1')
        else:
            # The latest release is the most recent non-prerelease,
            # non-draft release, sorted by the ``created_at`` attribute.
            # See: https://docs.github.com/en/free-pro-team@latest/rest/reference/repos#get-the-latest-release
            release_url = repo_url / 'releases/latest'

        async with self.manager.web_client.get(release_url) as response:
            if response.status == 404:
                raise R.PkgFileUnavailable('release not found')

            response_json = await response.json()
            if defn.strategy is Strategy.latest:
                (response_json,) = response_json
            release_metadata: _GithubRelease = response_json

        assets = release_metadata['assets']

        release_json = next(
            (a for a in assets if a['name'] == 'release.json' and a['state'] == 'uploaded'),
            None,
        )
        if release_json is None:
            logger.info(f'no release.json found for {defn}; inspecting assets')

            def is_valid_asset(asset: _GithubRelease_Asset):
                return (
                    # There is something of a convention that Classic archives
                    # end in '-classic' and lib-less archives end in '-nolib'.
                    # Archives produced by packager follow this convention
                    not asset['name'].endswith(('-nolib.zip', '-nolib-classic.zip'))
                    and asset['content_type']
                    in {'application/zip', 'application/x-zip-compressed'}
                    # A failed upload has a state value of 'starter'
                    and asset['state'] == 'uploaded'
                )

            matching_asset = next(
                chain(
                    (
                        a
                        for a in assets
                        if is_valid_asset(a)
                        and a['name'].endswith('-classic.zip')
                        is (self.manager.config.game_flavour is Flavour.vanilla_classic)
                    ),
                    filter(is_valid_asset, assets),
                ),
                None,
            )
            if matching_asset is None:
                raise R.PkgFileUnavailable

        else:
            logger.info(f'reading metadata for {defn} from release.json')

            packager_metadata: _PackagerReleaseJson = await manager.cache_response(
                self.manager, release_json['browser_download_url'], {'days': 1}
            )
            releases = packager_metadata['releases']
            if not releases:
                raise R.PkgFileUnavailable('no files available for download')

            matching_release = next(
                (
                    r
                    for r in releases
                    if r['nolib'] is False
                    and any(
                        m['flavor'] == self.release_json_flavours[self.manager.config.game_flavour]
                        for m in r['metadata']
                    )
                ),
                None,
            )
            if matching_release is None:
                raise R.PkgFileUnavailable(f'no files matching {self.manager.config.game_flavour}')

            matching_asset = next(
                (
                    a
                    for a in assets
                    if a['name'] == matching_release['filename'] and a['state'] == 'uploaded'
                ),
                None,
            )
            if matching_asset is None:
                raise R.PkgFileUnavailable(f'{matching_release["filename"]} not found')

        return models.Pkg(
            source=self.source,
            id=project_metadata['full_name'],
            slug=project_metadata['full_name'].lower(),
            name=project_metadata['name'],
            description=project_metadata['description'],
            url=project_metadata['html_url'],
            download_url=matching_asset['browser_download_url'],
            date_published=release_metadata['published_at'],
            version=release_metadata['tag_name'],
            changelog_url=_format_data_changelog(release_metadata['body']),
            options={'strategy': defn.strategy},
        )


class InstawowResolver(BaseResolver):
    source = 'instawow'
    name = 'instawow'
    strategies = frozenset({Strategy.default})
    changelog_format = ChangelogFormat.markdown

    _addons = {
        ('0', 'weakauras-companion'),
        ('1', 'weakauras-companion-autoupdate'),
    }

    async def resolve_one(self, defn: Defn, metadata: None) -> models.Pkg:
        try:
            source_id, slug = next(p for p in self._addons if defn.alias in p)
        except StopIteration:
            raise R.PkgNonexistent

        from .wa_updater import BuilderConfig, WaCompanionBuilder

        builder = WaCompanionBuilder(self.manager, BuilderConfig())
        if source_id == '1':
            await builder.build()

        return models.Pkg(
            source=self.source,
            id=source_id,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=builder.addon_zip_path.as_uri(),
            date_published=datetime.now(timezone.utc),
            version=(await builder.get_checksum())[:7],
            changelog_url=builder.changelog_path.as_uri(),
            options={'strategy': defn.strategy},
        )

    @classmethod
    async def catalogue(
        cls, web_client: _deferred_types.aiohttp.ClientSession
    ) -> AsyncIterator[CatatalogueBaseEntry]:
        yield CatatalogueBaseEntry(
            source=cls.source,
            id='1',
            slug='weakauras-companion-autoupdate',
            name='WeakAuras Companion',
            folders=[
                ['WeakAurasCompanion'],
            ],
            game_flavours=set(Flavour),
            download_count=1,
            last_updated=datetime.now(timezone.utc),
        )



from functools import lru_cache
from io import BytesIO
import json
import os
from pathlib import Path
import re
from zipfile import ZipFile

import pytest

from instawow import __version__
from instawow.config import Config, Flavour
from instawow.manager import Manager, init_web_client

inf = float('inf')

FIXTURES = Path(__file__).parent / 'fixtures'


def pytest_addoption(parser):
    parser.addoption('--iw-no-mock', action='store_true')


def should_mock(fn):
    import inspect
    import warnings

    def wrapper(request):
        if request.config.getoption('--iw-no-mock'):
            warnings.warn('not mocking')
            return None
        elif any(m.name == 'iw_no_mock' for m in request.node.iter_markers()):
            return None
        elif request.module.__name__ == 'test_json_rpc_api':
            # aresponses conflicts with aiohttp's own test server
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


def load_fixture(filename):
    return (FIXTURES / filename).read_bytes()


def load_json_fixture(filename):
    return json.loads(load_fixture(filename))


@lru_cache(maxsize=None)
def make_addon_zip(*folders):
    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        for folder in folders:
            file.writestr(f'{folder}/{folder}.toc', b'')

    return buffer.getvalue()


@pytest.fixture(scope='session', autouse=True)
def iw_temp_dir(tmp_path_factory):
    temp_dir = tmp_path_factory.mktemp('temp')
    os.environ['INSTAWOW_TEMP_DIR'] = str(temp_dir)
    yield temp_dir


@pytest.fixture(
    params=[
        Flavour.retail,
        Flavour.vanilla_classic,
    ]
)
def iw_config_dict_no_config_dir(tmp_path, request, iw_temp_dir):
    addons = tmp_path / 'wow' / 'interface' / 'addons'
    addons.mkdir(parents=True)
    return {
        'profile': '__default__',
        'addon_dir': addons,
        'temp_dir': iw_temp_dir,
        'game_flavour': request.param,
    }


@pytest.fixture
def iw_config_dict(tmp_path, iw_config_dict_no_config_dir):
    return {**iw_config_dict_no_config_dir, 'config_dir': tmp_path / 'config'}


@pytest.fixture
def iw_config(iw_config_dict):
    yield Config(**iw_config_dict).write()


@pytest.fixture
async def iw_web_client():
    async with init_web_client() as web_client:
        yield web_client


@pytest.fixture
def iw_manager(iw_config, iw_web_client):
    manager = Manager.from_config(iw_config)
    manager.contextualise(web_client=iw_web_client)
    yield manager


@pytest.fixture
@should_mock
def mock_pypi(aresponses):
    aresponses.add(
        'pypi.org',
        '/pypi/instawow/json',
        'get',
        {'info': {'version': __version__}},
        repeat=inf,
    )


@pytest.fixture
@should_mock
def mock_master_catalogue(aresponses):
    aresponses.add(
        'raw.githubusercontent.com',
        aresponses.ANY,
        'get',
        load_json_fixture('master-catalogue.json'),
        repeat=inf,
    )


@pytest.fixture
@should_mock
def mock_curse(aresponses, mock_master_catalogue):
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon',
        'post',
        load_json_fixture('curse-addon--all.json'),
        repeat=inf,
    )
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon/20338/files',
        'get',
        load_json_fixture('curse-addon-files.json'),
        repeat=inf,
    )
    aresponses.add(
        'addons-ecs.forgesvc.net',
        re.compile(r'^/api/v2/addon/20338/file/(\d+)/changelog'),
        'get',
        aresponses.Response(text=load_fixture('curse-addon-changelog.txt').decode()),
        repeat=inf,
    )
    aresponses.add(
        'edge.forgecdn.net',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_addon_zip('Molinari')),
        repeat=inf,
    )


@pytest.fixture
@should_mock
def mock_wowi(aresponses, mock_master_catalogue):
    aresponses.add(
        'api.mmoui.com',
        '/v3/game/WOW/filelist.json',
        'get',
        load_json_fixture('wowi-filelist.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.mmoui.com',
        re.compile(r'^/v3/game/WOW/filedetails/'),
        'get',
        load_json_fixture('wowi-filedetails.json'),
        repeat=inf,
    )
    aresponses.add(
        'cdn.wowinterface.com',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_addon_zip('Molinari')),
        repeat=inf,
    )


@pytest.fixture
@should_mock
def mock_tukui(aresponses, mock_master_catalogue):
    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=tukui',
        'get',
        load_json_fixture('tukui-ui--tukui.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=elvui',
        'get',
        load_json_fixture('tukui-ui--elvui.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?addons=all',
        'get',
        load_json_fixture('tukui-retail-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-addons=all',
        'get',
        load_json_fixture('tukui-classic-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-tbc-addons=all',
        'get',
        load_json_fixture('tukui-classic-tbc-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php',
        'get',
        '',
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        re.compile(r'^/downloads/tukui'),
        'get',
        aresponses.Response(body=make_addon_zip('Tukui')),
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/addons.php?download=1',
        'get',
        aresponses.Response(body=make_addon_zip('ElvUI_MerathilisUI')),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        re.compile(r'/classic-(?:-tbc)?addons\.php\?download=1'),
        'get',
        aresponses.Response(body=make_addon_zip('Tukui')),
        match_querystring=True,
        repeat=inf,
    )


@pytest.fixture
@should_mock
def mock_github(aresponses):
    aresponses.add(
        'api.github.com',
        '/repos/nebularg/PackagerTest',
        'get',
        load_json_fixture('github-repo-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/nebularg/PackagerTest/releases/latest',
        'get',
        load_json_fixture('github-release-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        '/nebularg/PackagerTest/releases/download/v1.9.6/release.json',
        'get',
        load_json_fixture('github-release-release-json-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras',
        'get',
        load_json_fixture('github-repo-legacy-lib-and-nolib.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/latest',
        'get',
        load_json_fixture('github-release-legacy-lib-and-nolib.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases?per_page=1',
        'get',
        [load_json_fixture('github-release-legacy-lib-and-nolib.json')],
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/tags/2.1.0',
        'get',
        load_json_fixture('github-release-legacy-lib-and-nolib-older-version.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/p3lim-wow/Molinari',
        'get',
        load_json_fixture('github-repo-legacy-retail-and-classic.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/p3lim-wow/Molinari/releases/latest',
        'get',
        load_json_fixture('github-release-legacy-retail-and-classic.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiBags',
        'get',
        load_json_fixture('github-repo-no-releases.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiBags/releases/latest',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/tags/2.0.19',
        'get',
        load_json_fixture('github-release-no-assets.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/layday/foobar',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        re.compile(r'^(/[^/]*){2}/releases/download'),
        'get',
        aresponses.Response(body=make_addon_zip('Foo')),
        repeat=inf,
    )


@pytest.fixture(autouse=True)
@should_mock
def mock_all(
    mock_pypi,
    mock_master_catalogue,
    mock_curse,
    mock_wowi,
    mock_tukui,
    mock_github,
):
    pass



from __future__ import annotations

from datetime import datetime
import typing
from typing import Any, Mapping

from pydantic import BaseModel
import sqlalchemy as sa
import sqlalchemy.future as sa_future
from typing_extensions import TypeAlias, TypeGuard

from . import db
from .common import Strategy


class _PkgOptions(BaseModel):
    strategy: Strategy


class _PkgFolder(BaseModel):
    name: str


class _PkgDep(BaseModel):
    id: str


class _PkgLoggedVersion(BaseModel):
    version: str
    install_time: datetime


class Pkg(BaseModel):
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: datetime
    version: str
    changelog_url: str
    options: _PkgOptions  # pkg_options
    folders: typing.List[_PkgFolder] = []  # pkg_folder
    deps: typing.List[_PkgDep] = []  # pkg_dep
    logged_versions: typing.List[_PkgLoggedVersion] = []  # pkg_version_log

    @classmethod
    def from_row_mapping(
        cls, connection: sa_future.Connection, row_mapping: Mapping[str, Any]
    ) -> Pkg:
        return cls.parse_obj(
            {
                **row_mapping,
                'folders': connection.execute(
                    sa.select(db.pkg_folder.c.name).filter_by(
                        pkg_source=row_mapping['source'], pkg_id=row_mapping['id']
                    )
                )
                .mappings()
                .all(),
                'options': connection.execute(
                    sa.select(db.pkg_options.c.strategy).filter_by(
                        pkg_source=row_mapping['source'], pkg_id=row_mapping['id']
                    )
                )
                .mappings()
                .one(),
                'deps': connection.execute(
                    sa.select(db.pkg_dep.c.id).filter_by(
                        pkg_source=row_mapping['source'], pkg_id=row_mapping['id']
                    )
                )
                .mappings()
                .all(),
                'logged_versions': connection.execute(
                    sa.select(db.pkg_version_log)
                    .filter_by(pkg_source=row_mapping['source'], pkg_id=row_mapping['id'])
                    .order_by(db.pkg_version_log.c.install_time.desc())
                    .limit(10)
                )
                .mappings()
                .all(),
            }
        )

    def insert(self, connection: sa_future.Connection) -> None:
        pkg_dict = self.dict()
        connection.execute(
            sa.insert(db.pkg),
            [pkg_dict],
        )
        connection.execute(
            sa.insert(db.pkg_folder),
            [
                {**f, 'pkg_source': pkg_dict['source'], 'pkg_id': pkg_dict['id']}
                for f in pkg_dict['folders']
            ],
        )
        connection.execute(
            sa.insert(db.pkg_options),
            [
                {
                    **pkg_dict['options'],
                    'pkg_source': pkg_dict['source'],
                    'pkg_id': pkg_dict['id'],
                }
            ],
        )
        if pkg_dict['deps']:
            connection.execute(
                sa.insert(db.pkg_dep),
                [
                    {**f, 'pkg_source': pkg_dict['source'], 'pkg_id': pkg_dict['id']}
                    for f in pkg_dict['deps']
                ],
            )
        connection.execute(
            sa.insert(db.pkg_version_log).prefix_with('OR IGNORE'),
            [
                {
                    'version': pkg_dict['version'],
                    'pkg_source': pkg_dict['source'],
                    'pkg_id': pkg_dict['id'],
                }
            ],
        )
        connection.commit()

    def delete(self, connection: sa_future.Connection) -> None:
        connection.execute(
            sa.delete(db.pkg_dep).filter_by(pkg_source=self.source, pkg_id=self.id),
        )
        connection.execute(
            sa.delete(db.pkg_folder).filter_by(pkg_source=self.source, pkg_id=self.id),
        )
        connection.execute(
            sa.delete(db.pkg_options).filter_by(pkg_source=self.source, pkg_id=self.id),
        )
        connection.execute(
            sa.delete(db.pkg).filter_by(source=self.source, id=self.id),
        )
        connection.commit()

    # Pydantic sets this to ``None`` unless the model is marked as "frozen",
    # presumably because it overrides ``__eq__``, which we're not interested in.
    __hash__ = object.__hash__


class PkgList(BaseModel):
    __root__: typing.List[Pkg]


def is_pkg(value: object) -> TypeGuard[Pkg]:
    return isinstance(value, Pkg)


PkgLike: TypeAlias = Any



from pathlib import Path
import sys

import pytest

from instawow.config import Config, Flavour


def test_env_vars_have_prio(iw_config_dict, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', '/foo')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    config = Config(**iw_config_dict)
    assert config.config_dir == Path('/foo').resolve()
    assert config.game_flavour is Flavour.burning_crusade_classic


def test_config_dir_is_populated(iw_config_dict):
    config = Config(**iw_config_dict).write()
    assert {i.name for i in config.profile_dir.iterdir()} == {'config.json', 'logs', 'plugins'}


def test_reading_missing_config_from_env_raises(iw_config_dict, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_config_dict['config_dir']))
    with pytest.raises(FileNotFoundError):
        Config.read('__default__')


def test_missing_addon_dir_raises(iw_config_dict):
    with pytest.raises(ValueError):
        Config(**{**iw_config_dict, 'addon_dir': 'foo'})


@pytest.mark.skipif(sys.platform == 'win32', reason='path handling')
def test_default_config_dir_is_platform_appropriate(iw_config_dict_no_config_dir, monkeypatch):
    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')
        config_dir = Config(**iw_config_dict_no_config_dir).config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = Config(**iw_config_dict_no_config_dir).config_dir
        assert config_dir == Path('/foo/instawow')

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')
        config_dir = Config(**iw_config_dict_no_config_dir).config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'


@pytest.mark.skipif(sys.platform != 'win32', reason='path unhandling')
def test_default_config_dir_is_win32_appropriate(iw_config_dict_no_config_dir, monkeypatch):
    assert (
        Config(**iw_config_dict_no_config_dir).config_dir
        == Path.home() / 'AppData/Roaming/instawow'
    )
    monkeypatch.delenv('APPDATA')
    assert Config(**iw_config_dict_no_config_dir).config_dir == Path.home() / 'instawow'


def test_can_infer_flavour_from_path():
    assert (
        Config.infer_flavour('wowzerz/_classic_/Interface/AddOns')
        is Flavour.burning_crusade_classic
    )
    assert (
        Config.infer_flavour('/foo/bar/_classic_beta_/Interface/AddOns')
        is Flavour.burning_crusade_classic
    )
    assert (
        Config.infer_flavour('/foo/bar/_classic_ptr_/Interface/AddOns')
        is Flavour.burning_crusade_classic
    )
    assert Config.infer_flavour('_classic_era_/Interface/AddOns') is Flavour.vanilla_classic
    assert Config.infer_flavour('_classic_era_ptr_/Interface/AddOns') is Flavour.vanilla_classic
    assert Config.infer_flavour('wowzerz/_retail_/Interface/AddOns') is Flavour.retail
    assert Config.infer_flavour('anything goes') is Flavour.retail


def test_can_list_profiles(monkeypatch, iw_config_dict):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_config_dict['config_dir']))
    assert Config.list_profiles() == []
    Config.parse_obj(iw_config_dict).write()
    Config.parse_obj({**iw_config_dict, 'profile': 'foo'}).write()
    assert sorted(Config.list_profiles()) == ['__default__', 'foo']


def test_can_delete_profile(iw_config_dict):
    config = Config(**iw_config_dict).write()
    assert config.profile_dir.exists()
    config.delete()
    assert not config.profile_dir.exists()



from __future__ import annotations

import asyncio
from itertools import product
from pathlib import Path
import sys
import time

import pytest

from instawow.manager import find_addon_zip_base_dirs, make_zip_member_filter
from instawow.utils import (
    TocReader,
    bucketise,
    file_uri_to_path,
    merge_intersecting_sets,
    run_in_thread,
    tabulate,
)


@pytest.fixture
def fake_addon():
    yield Path(__file__).parent / 'fixtures' / 'FakeAddon'


def test_find_addon_zip_base_dirs_can_find_explicit_dirs():
    assert set(find_addon_zip_base_dirs(['b/', 'b/b.toc'])) == {'b'}


def test_find_addon_zip_base_dirs_can_find_implicit_dirs():
    assert set(find_addon_zip_base_dirs(['b/b.toc'])) == {'b'}


def test_find_addon_zip_base_dirs_discards_tocless_paths():
    assert set(find_addon_zip_base_dirs(['a', 'b/b.toc', 'c/'])) == {'b'}


def test_find_addon_zip_base_dirs_discards_mismatched_tocs():
    assert not set(find_addon_zip_base_dirs(['a', 'a/b.toc']))


@pytest.mark.parametrize('ext', product('Tt', 'Oo', 'Cc'))
def test_find_addon_zip_base_dirs_toc_is_case_insensitive(ext: tuple[str, ...]):
    assert set(find_addon_zip_base_dirs([f'a/a.{"".join(ext)}'])) == {'a'}


def test_make_zip_member_filter_discards_names_with_prefix_not_in_dirs():
    is_member = make_zip_member_filter({'b'})
    assert list(filter(is_member, ['a/', 'b/', 'aa/', 'bb/', 'b/c', 'a/d'])) == ['b/', 'b/c']


def test_loading_toc_from_addon_path(fake_addon):
    TocReader.from_addon_path(fake_addon)
    with pytest.raises(FileNotFoundError):
        TocReader.from_addon_path(fake_addon.parent / 'MissingAddon')


def test_parsing_toc_entries(fake_addon):
    toc_reader = TocReader.from_addon_path(fake_addon)
    assert toc_reader.entries == {
        'Normal': 'Normal entry',
        'Compact': 'Compact entry',
    }


def test_toc_entry_indexing(fake_addon):
    toc_reader = TocReader.from_addon_path(fake_addon)
    assert toc_reader['Normal'] == 'Normal entry'
    assert toc_reader['Compact'] == 'Compact entry'
    assert toc_reader['Indented'] is None
    assert toc_reader['Comment'] is None
    assert toc_reader['Nonexistent'] is None


def test_toc_entry_multiindexing(fake_addon):
    toc_reader = TocReader.from_addon_path(fake_addon)
    assert toc_reader['Normal', 'Compact'] == 'Normal entry'
    assert toc_reader['Compact', 'Normal'] == 'Compact entry'
    assert toc_reader['Indented', 'Normal'] == 'Normal entry'
    assert toc_reader['Nonexistent', 'Indented'] is None
    assert toc_reader['Nonexistent', 'Indented', 'Normal'] == 'Normal entry'


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


def test_tabulate_spits_out_ascii_table(fake_addon):
    toc_reader = TocReader.from_addon_path(fake_addon)
    data = [('key', 'value'), *toc_reader.entries.items()]
    assert tabulate(data) == (
        '  key        value    \n'
        '-------  -------------\n'
        'Normal   Normal entry \n'
        'Compact  Compact entry'
    )


def test_merge_intersecting_sets_in_noncontiguous_collection():
    collection = [
        {'a'},
        {'b', 'c'},
        {'a', 'd'},
        {'e'},
        {'c', 'f'},
        {'g', 'a'},
    ]
    output = [
        {'a', 'd', 'g'},
        {'b', 'c', 'f'},
        {'e'},
    ]
    assert sorted(merge_intersecting_sets(collection)) == output


@pytest.mark.skipif(sys.platform == 'win32', reason='platform dependent')
def test_file_uri_to_path_posix_leading_slash_is_preserved():
    uri = Path('/foo/bar').as_uri()
    assert uri == 'file:///foo/bar'
    assert file_uri_to_path(uri) == '/foo/bar'


@pytest.mark.skipif(sys.platform != 'win32', reason='platform dependent')
def test_file_uri_to_path_win32_leading_slash_is_stripped():
    uri = Path('C:/foo/bar').as_uri()
    assert uri == 'file:///C:/foo/bar'
    assert file_uri_to_path(uri) == 'C:/foo/bar'


@pytest.mark.asyncio
async def test_generator_in_run_in_thread_does_not_lock_up_loop():
    def foo():
        time.sleep(2)
        yield 'foo'

    async def bar():
        await asyncio.sleep(1)
        return ['bar']

    assert [await a for a in asyncio.as_completed([run_in_thread(list)(foo()), bar()])] == [
        ['bar'],
        ['foo'],
    ]



from datetime import datetime

import click

from instawow.common import Strategy
from instawow.models import Pkg
import instawow.plugins
from instawow.resolvers import BaseResolver, ChangelogFormat, Defn


@click.command()
def foo():
    "don't foo where you bar"
    print('success!')


class MyResolver(BaseResolver):
    source = 'me'
    name = "It's me"
    strategies = frozenset({Strategy.default})
    changelog_format = ChangelogFormat.markdown

    async def resolve_one(self, defn: Defn, metadata: None) -> Pkg:
        return Pkg(
            source=self.source,
            id='bar',
            slug='bar',
            name='Bar',
            description='The quintessential bar add-on, brought to you by yours truly',
            url='http://example.com/',
            download_url='file:///...',
            date_published=datetime.now(),
            version='0',
            changelog_url='data:,',
            options={'strategy': defn.strategy},
        )


@instawow.plugins.hookimpl
def instawow_add_commands():
    return (foo,)


@instawow.plugins.hookimpl
def instawow_add_resolvers():
    return (MyResolver,)



# These tests are adapted from SLPP <https://github.com/SirAnthony/slpp>.
#
#   Copyright (c) 2010, 2011, 2012 SirAnthony <anthony at adsorbtion.org>
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.

import pytest

from instawow._custom_slpp import SLPP, ParseError


def decode(value):
    return SLPP(value).decode()


def test_numbers():
    # int and float
    assert decode('3') == 3
    assert decode('3.') == 3
    assert decode('3.1') == 3.1
    # negative float
    assert decode('-0.45') == -0.45
    # scientific
    assert decode('3e-07') == 3e-7
    assert decode('-3.23e+17') == -3.23e17
    # hex
    assert decode('0x3a') == 0x3A
    assert (
        decode(
            '''{
                ID = 0x74fa4cae,
                Version = 0x07c2,
                Manufacturer = 0x21544948
            }'''
        )
        == {'ID': 0x74FA4CAE, 'Version': 0x07C2, 'Manufacturer': 0x21544948}
    )


def test_bool():
    assert decode('true') == True
    assert decode('false') == False
    assert decode('falser') == 'falser'


def test_bool_keys_are_not_nums():
    assert decode('{ [false] = "", [true] = "" }') == {False: '', True: ''}
    assert decode('{ [false] = "", [true] = "", [3] = "" }') == {False: '', True: '', 3: ''}


def test_nil():
    decode('nil') == None


def test_simple_table():
    # bracketed key
    assert decode('{ [10] = 11 }') == {10: 11}
    assert decode('{ [false] = 0 }') == {False: 0}
    assert decode('{ [true] = 1 }') == {True: 1}
    # bracketed string key - not supported
    # assert decode('{ [ [[10]] ] = 1 }') == {"10": 1}
    # keywords in table
    assert decode('{ false, true }') == [False, True]
    # void table
    assert decode('{ nil }') == {}
    # values-only table
    assert decode('{ "10" }') == ["10"]
    # last zero
    assert decode('{ 0, 1, 0 }') == [0, 1, 0]


def test_string_with_and_without_escape():
    assert decode("'test\\'s string'") == "test's string"
    assert decode('"test\\\'s string"') == "test\\'s string"
    assert decode('"test\'s string"') == "test's string"
    assert decode("[[test\\'s string]]") == "test\\'s string"
    assert decode("[[test's string]]") == "test's string"
    # https://github.com/SirAnthony/slpp/issues/23
    assert decode("'--3'") == '--3'


def test_table_keys():
    assert decode('{ [ [[false]] ] = "" }') == {'false': ''}
    with pytest.raises(ParseError):
        decode('{ [[false]] = "" }')


def test_table_palooza():
    assert decode(
        '''{ -- δκσξδφξ
        array = { 65, 23, 5 }, -- 3493
        dict =     {  -- !!!,11
            [false]       =    "value"      ,  -- what's up
            ["array"] = {  -- waddup
                3, [[6]],   -- wassup
                4 }, -- [2]
            mixed = { 43, 54.3, false, string = "value", 9 }    -- wazzup
        }                   -- foo
} -- bar'''
    ) == {
        'array': [65, 23, 5],
        'dict': {
            False: 'value',
            'array': [3, '6', 4],
            'mixed': {1: 43, 2: 54.3, 3: False, 'string': 'value', 4: 9},
        },
    }


def test_table_key_overrides():
    assert decode(
        '{ 43, 54.3, false, string = "value", 9, [4] = 111, [1] = 222, [2.1] = "text" }'
    ) == {
        1: 43,
        2: 54.3,
        3: False,
        4: 9,
        'string': 'value',
        2.1: 'text',
    }
    assert decode('{ 43, 54.3, false, 9, [5] = 111, [7] = 222 }') == {
        1: 43,
        2: 54.3,
        3: False,
        4: 9,
        5: 111,
        7: 222,
    }
    assert decode('{ [7] = 111, [5] = 222, 43, 54.3, false, 9 }') == {
        7: 111,
        5: 222,
        1: 43,
        2: 54.3,
        3: False,
        4: 9,
    }
    assert decode('{ 43, 54.3, false, 9, [4] = 111, [5] = 52.1 }') == [
        43,
        54.3,
        False,
        9,
        52.1,
    ]
    assert decode('{ [5] = 111, [4] = 52.1, 43, [3] = 54.3, false, 9 }') == {
        5: 111,
        4: 52.1,
        1: 43,
        2: False,
        3: 9,
    }
    assert decode('{ [1] = 1, [2] = "2", 3, 4, [5] = 5, [5] = 6 }') == {
        1: 3,
        2: 4,
        5: 6,
    }



import datetime
from pathlib import Path

from aiohttp import ClientError
import pytest

from instawow import results as R
from instawow.common import Strategy
from instawow.config import Flavour
from instawow.manager import is_outdated
from instawow.models import Pkg
from instawow.resolvers import Defn


@pytest.mark.asyncio
async def test_pinning_supported_pkg(iw_manager):
    defn = Defn('curse', 'molinari')
    install_result = await iw_manager.install([defn], False)
    pkg = install_result[defn].pkg
    version = pkg.version

    for new_defn in (defn.with_version(pkg.version), defn):
        pin_result = await iw_manager.pin([new_defn])
        pinned_pkg = pin_result[new_defn].pkg
        assert pkg.options.strategy is Strategy.default
        assert pinned_pkg.options.strategy is new_defn.strategy
        assert version == pinned_pkg.version


@pytest.mark.asyncio
async def test_pinning_unsupported_pkg(iw_manager):
    molinari_defn = Defn('wowi', '13188')
    await iw_manager.install([molinari_defn], False)
    installed_pkg = iw_manager.get_pkg(molinari_defn)
    assert installed_pkg.options.strategy == Strategy.default
    result = await iw_manager.pin([molinari_defn])
    assert (
        type(result[molinari_defn]) is R.PkgStrategyUnsupported
        and result[molinari_defn].strategy is Strategy.version
    )
    assert installed_pkg.options.strategy == Strategy.default


@pytest.mark.asyncio
async def test_pinning_nonexistent_pkg(iw_manager):
    molinari_defn = Defn('wowi', '13188')
    result = await iw_manager.pin([molinari_defn])
    assert type(result[molinari_defn]) is R.PkgNotInstalled


@pytest.mark.parametrize('exception', [ValueError('foo'), ClientError('bar')])
@pytest.mark.asyncio
async def test_resolve_rewraps_exception_appropriately_from_resolve(
    monkeypatch, iw_manager, exception
):
    async def resolve_one(self, defn, metadata):
        raise exception

    monkeypatch.setattr('instawow.resolvers.CurseResolver.resolve_one', resolve_one)

    defn = Defn('curse', 'molinari')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is R.InternalError
    assert results[defn].message == f'internal error: "{exception}"'


@pytest.mark.parametrize('exception', [ValueError('foo'), ClientError('bar')])
@pytest.mark.asyncio
async def test_resolve_rewraps_exception_appropriately_from_batch_resolve(
    monkeypatch, iw_manager, exception
):
    async def resolve(self, defns):
        raise exception

    monkeypatch.setattr('instawow.resolvers.CurseResolver.resolve', resolve)

    defn = Defn('curse', 'molinari')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is R.InternalError
    assert results[defn].message == f'internal error: "{exception}"'


@pytest.mark.asyncio
async def test_resolve_invalid_source(iw_manager):
    defn = Defn('bar', 'baz')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is R.PkgSourceInvalid


@pytest.mark.asyncio
async def test_resolve_plugin_hook_source(iw_manager):
    pytest.importorskip('instawow_test_plugin')
    defn = Defn('me', 'bar')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is Pkg


@pytest.mark.asyncio
async def test_install_can_replace_unreconciled_folders(iw_manager):
    molinari = iw_manager.config.addon_dir / 'Molinari'
    molinari.mkdir()

    defn = Defn('curse', 'molinari')

    result = await iw_manager.install([defn], replace=False)
    assert type(result[defn]) is R.PkgConflictsWithUnreconciled
    assert not any(molinari.iterdir())

    result = await iw_manager.install([defn], replace=True)
    assert type(result[defn]) is R.PkgInstalled
    assert any(molinari.iterdir())


@pytest.mark.asyncio
async def test_install_cannot_replace_reconciled_folders(iw_manager):
    curse_defn = Defn('curse', 'molinari')
    wowi_defn = Defn('wowi', '13188-molinari')

    result = await iw_manager.install([curse_defn], replace=False)
    assert type(result[curse_defn]) is R.PkgInstalled

    result = await iw_manager.install([wowi_defn], replace=False)
    assert type(result[wowi_defn]) is R.PkgConflictsWithInstalled

    result = await iw_manager.install([wowi_defn], replace=True)
    assert type(result[wowi_defn]) is R.PkgConflictsWithInstalled


@pytest.mark.asyncio
async def test_update_lifecycle_while_varying_retain_defn_strategy(iw_manager):
    defn = Defn('curse', 'molinari')
    versioned_defn = defn.with_version('80000.57-Release')

    result = await iw_manager.install([defn], replace=False)
    assert type(result[defn]) is R.PkgInstalled
    assert result[defn].pkg.options.strategy == Strategy.default

    result = await iw_manager.update([defn], retain_defn_strategy=False)
    assert type(result[defn]) is R.PkgUpToDate
    assert result[defn].is_pinned is False

    result = await iw_manager.update([versioned_defn], retain_defn_strategy=False)
    assert type(result[versioned_defn]) is R.PkgUpToDate
    assert result[versioned_defn].is_pinned is False

    result = await iw_manager.update([versioned_defn], retain_defn_strategy=True)
    assert type(result[versioned_defn]) is R.PkgUpdated
    assert result[versioned_defn].new_pkg.options.strategy == Strategy.version

    result = await iw_manager.update([defn], retain_defn_strategy=False)
    assert type(result[defn]) is R.PkgUpToDate
    assert result[defn].is_pinned is True

    result = await iw_manager.update([defn], retain_defn_strategy=True)
    assert type(result[defn]) is R.PkgUpdated
    assert result[defn].new_pkg.options.strategy == Strategy.default


@pytest.mark.parametrize('keep_folders', [True, False])
@pytest.mark.asyncio
async def test_deleting_and_retaining_folders_on_remove(iw_manager, keep_folders):
    defn = Defn('curse', 'molinari')

    await iw_manager.install([defn], False)
    folders = [iw_manager.config.addon_dir / f.name for f in iw_manager.get_pkg(defn).folders]
    assert all(f.is_dir() for f in folders)

    result = await iw_manager.remove([defn], keep_folders=keep_folders)
    assert type(result[defn]) is R.PkgRemoved
    assert not iw_manager.get_pkg(defn)
    if keep_folders:
        assert all(f.is_dir() for f in folders)
    else:
        assert not any(f.is_dir() for f in folders)


@pytest.mark.parametrize('keep_folders', [True, False])
@pytest.mark.asyncio
async def test_removing_pkg_with_missing_folders(iw_manager, keep_folders):
    defn = Defn('curse', 'molinari')

    result = await iw_manager.install([defn], False)
    folders = [iw_manager.config.addon_dir / f.name for f in result[defn].pkg.folders]
    for folder in folders:
        folder.rename(folder.with_name(f'Not_{folder.name}'))
    assert not any(f.is_dir() for f in folders)

    result = await iw_manager.remove([defn], keep_folders=keep_folders)
    assert type(result[defn]) is R.PkgRemoved
    assert not iw_manager.get_pkg(defn)


@pytest.mark.asyncio
async def test_basic_search(iw_manager):
    limit = 5
    results = await iw_manager.search('molinari', limit=limit)
    assert len(results) == limit
    assert {('curse', 'molinari'), ('wowi', '13188')} <= {
        (e.source, e.slug or e.id) for e in results
    }


@pytest.mark.asyncio
async def test_search_flavour_filtering(iw_manager):
    results = await iw_manager.search('AtlasLootClassic', limit=5)
    faux_defns = {(e.source, e.slug or e.id) for e in results}
    if iw_manager.config.game_flavour is Flavour.vanilla_classic:
        assert ('curse', 'atlaslootclassic') in faux_defns
    else:
        assert ('curse', 'atlaslootclassic') not in faux_defns


@pytest.mark.asyncio
async def test_search_source_filtering(iw_manager):
    results = await iw_manager.search('molinari', limit=5, sources={'curse'})
    assert all(e.source == 'curse' for e in results)
    assert {('curse', 'molinari')} <= {(e.source, e.slug) for e in results}


@pytest.mark.asyncio
async def test_search_date_filtering(iw_manager):
    start_date = datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=365)

    results = await iw_manager.search('molinari', limit=5)
    assert any(e.last_updated < start_date for e in results)

    results = await iw_manager.search('molinari', limit=5, start_date=start_date)
    assert all(e.last_updated > start_date for e in results)


@pytest.mark.asyncio
async def test_search_unknown_source(iw_manager):
    with pytest.raises(ValueError, match='Unknown source'):
        await iw_manager.search('molinari', limit=5, sources={'foo'})


@pytest.mark.asyncio
async def test_get_changelog_from_empty_data_url(iw_manager):
    assert (await iw_manager.get_changelog('data:,')) == ''


@pytest.mark.asyncio
async def test_get_changelog_from_url_encoded_data_url(iw_manager):
    assert (await iw_manager.get_changelog('data:,foo%20bar')) == 'foo bar'


@pytest.mark.asyncio
async def test_get_malformed_changelog(iw_manager):
    with pytest.raises(ValueError, match='Unsupported URL with scheme'):
        await iw_manager.get_changelog('')


@pytest.mark.asyncio
async def test_get_changelog_from_file_uri(iw_manager):
    assert (
        await iw_manager.get_changelog(
            (Path(__file__).parent / 'fixtures' / 'curse-addon-changelog.txt').as_uri()
        )
    ).startswith('<h3>Changes in 90100.79-Release:</h3>')


@pytest.mark.asyncio
async def test_get_changelog_from_web_url(iw_manager):
    assert (
        await iw_manager.get_changelog(
            'https://addons-ecs.forgesvc.net/api/v2/addon/20338/file/3475338/changelog'
        )
    ).startswith('<h3>Changes in 90100.79-Release:</h3>')


@pytest.mark.iw_no_mock
@pytest.mark.asyncio
async def test_is_outdated_works_in_variety_of_scenarios(monkeypatch, aresponses, iw_temp_dir):
    pypi_version = iw_temp_dir.joinpath('.pypi_version')
    if pypi_version.exists():
        pypi_version.unlink()

    # version == '0.0.0', version not cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.0.0')
        assert await is_outdated() == (False, '')

    # Update check disabled, version not cached
    with monkeypatch.context() as patcher:
        patcher.setenv('INSTAWOW_AUTO_UPDATE_CHECK', '0')
        assert await is_outdated() == (False, '')

    # Endpoint not responsive, version not cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/pypi/instawow/json',
            'get',
            aresponses.Response(status=500),
        )
        assert await is_outdated() == (False, '0.1.0')

    # Endpoint responsive, version not cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/pypi/instawow/json',
            'get',
            {'info': {'version': '1.0.0'}},
        )
        assert await is_outdated() == (True, '1.0.0')

    # version == '0.0.0', version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.0.0')
        assert await is_outdated() == (False, '')

    # Update check disabled, version cached
    with monkeypatch.context() as patcher:
        patcher.setenv('INSTAWOW_AUTO_UPDATE_CHECK', '0')
        assert await is_outdated() == (False, '')

    # Endpoint not responsive, version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/pypi/instawow/json',
            'get',
            aresponses.Response(status=500),
        )
        assert await is_outdated() == (True, '1.0.0')

    # Endpoint responsive, version cached and version same
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '0.1.0')
        aresponses.add(
            'pypi.org',
            '/pypi/instawow/json',
            'get',
            {'info': {'version': '1.0.0'}},
        )
        assert await is_outdated() == (True, '1.0.0')

    # Endpoint responsive, version cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow.__version__', '1.0.0')
        aresponses.add(
            'pypi.org',
            '/pypi/instawow/json',
            'get',
            {'info': {'version': '1.0.0'}},
        )
        assert await is_outdated() == (False, '1.0.0')



from functools import partial
import json

from aiohttp.test_utils import TestClient, TestServer
import pytest

from instawow.config import Config

try:
    from instawow_gui import json_rpc_server
except ImportError:
    pytestmark = pytest.mark.skip(reason='instawow_gui is not available')


dumps = partial(json.dumps, default=str)


@pytest.fixture
async def ws(iw_config_dict, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_config_dict['config_dir']))
    app = await json_rpc_server.create_app()
    server = TestServer(app)
    async with TestClient(server) as client, client.ws_connect('/api') as ws:
        yield ws


@pytest.mark.asyncio
async def test_write_config(request, iw_config_dict, ws):
    config_values = {**iw_config_dict, 'profile': request.node.name}
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'config/write',
        'params': {'values': config_values, 'infer_game_flavour': False},
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['id'] == request.node.name
    assert Config.parse_obj(rpc_response['result']) == Config.parse_obj(config_values)


@pytest.mark.asyncio
async def test_write_config_with_invalid_params(request, iw_config_dict, ws):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'config/write',
        'params': {
            'values': {**iw_config_dict, 'game_flavour': 'strawberry'},
            'infer_game_flavour': False,
        },
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['id'] == request.node.name
    assert rpc_response['error']
    assert rpc_response['error']['code'] == -32001
    assert rpc_response['error']['message'] == 'invalid configuration parameters'
    assert rpc_response['error']['data'] == [
        {
            'loc': ['game_flavour'],
            'msg': "value is not a valid enumeration member; permitted: 'retail', 'vanilla_classic', 'classic'",
            'type': 'type_error.enum',
            'ctx': {'enum_values': ['retail', 'vanilla_classic', 'classic']},
        }
    ]


@pytest.mark.asyncio
async def test_install_with_invalid_params(request, ws):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'install',
        'params': {},
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['error'] and rpc_response['error']['code'] == -32602


@pytest.mark.xfail
@pytest.mark.asyncio
async def test_install_with_uninitialised_profile(request, ws):
    rpc_request = {
        'jsonrpc': '2.0',
        'method': 'install',
        'params': {
            'profile': request.node.name,
            'defns': [{'source': 'curse', 'name': 'molinari'}],
            'replace': False,
        },
        'id': request.node.name,
    }
    await ws.send_json(rpc_request, dumps=dumps)
    rpc_response = await ws.receive_json()
    assert rpc_response['error'] and rpc_response['error']['code'] == -32001



from __future__ import annotations

import atexit
import sys
import threading

from cefpython3 import cefpython as cef
from toga_winforms.widgets.box import Box

_cef_loaded = False


def load() -> None:
    global _cef_loaded

    if not _cef_loaded:
        sys.excepthook = cef.ExceptHook
        cef.Initialize(
            settings={
                'cache_path': '',
                'multi_threaded_message_loop': True,
            },
            switches={
                # To enable ``filter-backdrop`` support
                'enable-experimental-web-platform-features': '',
            },
        )
        atexit.register(cef.Shutdown)
        _cef_loaded = True


class _CefWidget(Box):
    def create(self):
        super().create()

        self.window_created = threading.Event()
        window_info = cef.WindowInfo()
        window_info.SetAsChild(self.native.Handle.ToInt32())

        def create_browser(widget: _CefWidget, window_info: cef.WindowInfo) -> None:
            widget.browser = cef.CreateBrowserSync(window_info)
            widget.browser.WasResized()
            self.window_created.set()

        cef.PostTask(cef.TID_UI, create_browser, self, window_info)

    def set_bounds(self, x: int, y: int, width: int, height: int) -> None:
        super().set_bounds(x, y, width, height)

        def update_window_size():
            cef.WindowUtils.OnSize(self.native.Handle.ToInt32(), 0, 0, 0)

        cef.PostTask(cef.TID_UI, update_window_size)

    def set_on_key_down(self, handler: object) -> None:
        pass

    def set_on_webview_load(self, handler: object) -> None:
        pass

    def set_url(self, value: str) -> None:
        def load_url():
            self.window_created.wait()
            self.browser.LoadUrl(value)

        cef.PostTask(cef.TID_UI, load_url)

    def set_content(self, root_url: str, content: str) -> None:
        raise NotImplementedError

    def set_user_agent(self, value: str) -> None:
        pass

    async def evaluate_javascript(self, javascript: str) -> None:
        raise NotImplementedError

    def invoke_javascript(self, javascript: str) -> None:
        def execute_javascript():
            self.window_created.wait()
            self.browser.ExecuteJavascript(javascript)

        cef.PostTask(cef.TID_UI, execute_javascript)


class Factory:
    WebView = _CefWidget



from functools import partial
import json
from textwrap import dedent

from click.testing import CliRunner
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput
import pytest

from instawow import __version__
from instawow.cli import main
from instawow.config import Flavour
from instawow.manager import Manager
from instawow.models import PkgList


@pytest.fixture
def feed_pt():
    pipe_input = create_pipe_input()
    with create_app_session(input=pipe_input, output=DummyOutput()):
        yield pipe_input.send_text
    pipe_input.close()


@pytest.fixture
def run(monkeypatch, event_loop, iw_config, iw_web_client):
    def runner(awaitable):
        Manager.contextualise(web_client=iw_web_client)
        return event_loop.run_until_complete(awaitable)

    monkeypatch.setattr('instawow.cli.run_with_progress', runner)
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_config.config_dir))
    yield partial(CliRunner().invoke, main, catch_exceptions=False)


@pytest.fixture
def molinari_and_run(run):
    run('install curse:molinari')
    yield run


@pytest.fixture
def faux_molinari_and_run(iw_config, run):
    molinari = iw_config.addon_dir / 'Molinari'
    molinari.mkdir()
    (molinari / 'Molinari.toc').write_text(
        '''\
## X-Curse-Project-ID: 20338
## X-WoWI-ID: 13188
'''
    )
    yield run


def test_valid_curse_pkg_lifecycle(iw_config, run):
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert (
        run('install curse:molinari').output == '✗ curse:molinari\n  package already installed\n'
    )
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is up to date\n'
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is not installed\n'
    assert run('remove curse:molinari').output == '✗ curse:molinari\n  package is not installed\n'


def test_valid_tukui_pkg_lifecycle(iw_config, run):
    assert run('install tukui:1').output.startswith('✓ tukui:1\n  installed')
    assert run('install tukui:1').output == '✗ tukui:1\n  package already installed\n'
    assert run('update tukui:1').output == '✗ tukui:1\n  package is up to date\n'
    assert run('remove tukui:1').output == '✓ tukui:1\n  removed\n'
    assert run('update tukui:1').output == '✗ tukui:1\n  package is not installed\n'
    assert run('remove tukui:1').output == '✗ tukui:1\n  package is not installed\n'
    if iw_config.game_flavour is Flavour.retail:
        assert run('install tukui:-1').output.startswith('✓ tukui:-1\n  installed')
        assert run('install tukui:-1').output == '✗ tukui:-1\n  package already installed\n'
        assert run('update tukui:-1').output == '✗ tukui:-1\n  package is up to date\n'
        assert run('remove tukui:-1').output == '✓ tukui:-1\n  removed\n'
    else:
        assert run('install tukui:-1').output == '✗ tukui:-1\n  package does not exist\n'
    assert run('update tukui:-1').output == '✗ tukui:-1\n  package is not installed\n'
    assert run('remove tukui:-1').output == '✗ tukui:-1\n  package is not installed\n'


def test_valid_wowi_pkg_lifecycle(run):
    assert run('install wowi:13188-molinari').output.startswith(
        '✓ wowi:13188-molinari\n  installed'
    )
    assert (
        run('install wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  package already installed\n'
    )
    assert (
        run('update wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  package is up to date\n'
    )
    assert run('remove wowi:13188-molinari').output == '✓ wowi:13188-molinari\n  removed\n'
    assert (
        run('update wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  package is not installed\n'
    )
    assert (
        run('remove wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  package is not installed\n'
    )


def test_invalid_source_lifecycle(run):
    assert run('install foo:bar').output == '✗ foo:bar\n  package source is invalid\n'
    assert run('update foo:bar').output == '✗ foo:bar\n  package is not installed\n'
    assert run('remove foo:bar').output == '✗ foo:bar\n  package is not installed\n'


def test_nonexistent_addon_alias_lifecycle(run):
    assert (
        run('install curse:gargantuan-wigs').output
        == '✗ curse:gargantuan-wigs\n  package does not exist\n'
    )
    assert (
        run('update curse:gargantuan-wigs').output
        == '✗ curse:gargantuan-wigs\n  package is not installed\n'
    )
    assert (
        run('remove curse:gargantuan-wigs').output
        == '✗ curse:gargantuan-wigs\n  package is not installed\n'
    )


def test_reconciled_folder_conflict_on_install(run):
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert run('install wowi:13188-molinari').output == (
        '✗ wowi:13188-molinari\n'
        '  package folders conflict with installed package Molinari\n'
        '    (curse:20338)\n'
    )


def test_unreconciled_folder_conflict_on_install(iw_config, run):
    iw_config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == "✗ curse:molinari\n  package folders conflict with 'Molinari'\n"
    )
    assert run('install --replace curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed'
    )


def test_keep_folders_on_remove(iw_config, molinari_and_run):
    assert (
        molinari_and_run('remove --keep-folders curse:molinari').output
        == '✓ curse:molinari\n  removed\n'
    )
    assert iw_config.addon_dir.joinpath('Molinari').is_dir()


def test_install_with_curse_alias(run):
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert run('install curse:20338').output == '✗ curse:20338\n  package already installed\n'
    assert (
        run('install https://www.wowace.com/projects/molinari').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert (
        run('install https://www.curseforge.com/wow/addons/molinari').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert (
        run('install https://www.curseforge.com/wow/addons/molinari/download').output
        == '✗ curse:molinari\n  package already installed\n'
    )


def test_install_with_tukui_alias(iw_config, run):
    if iw_config.game_flavour is Flavour.retail:
        assert run('install tukui:-1').output.startswith('✓ tukui:-1\n  installed')
        assert run('install tukui:tukui').output == '✗ tukui:tukui\n  package already installed\n'
        assert (
            run('install https://www.tukui.org/download.php?ui=tukui').output
            == '✗ tukui:tukui\n  package already installed\n'
        )
    else:
        assert run('install tukui:-1').output == '✗ tukui:-1\n  package does not exist\n'
        assert run('install tukui:tukui').output == '✗ tukui:tukui\n  package does not exist\n'
        assert (
            run('install https://www.tukui.org/download.php?ui=tukui').output
            == '✗ tukui:tukui\n  package does not exist\n'
        )
    assert run('install https://www.tukui.org/addons.php?id=1').output.startswith(
        '✓ tukui:1\n  installed'
    )
    assert run('install tukui:1').output == '✗ tukui:1\n  package already installed\n'


def test_install_with_wowi_alias(run):
    assert run('install wowi:13188').output.startswith('✓ wowi:13188\n  installed')
    assert (
        run('install wowi:13188-molinari').output
        == '✗ wowi:13188-molinari\n  package already installed\n'
    )
    assert (
        run('install https://www.wowinterface.com/downloads/landing.php?fileid=13188').output
        == '✗ wowi:13188\n  package already installed\n'
    )
    assert (
        run('install https://www.wowinterface.com/downloads/fileinfo.php?id=13188').output
        == '✗ wowi:13188\n  package already installed\n'
    )
    assert (
        run('install https://www.wowinterface.com/downloads/download13188-Molinari').output
        == '✗ wowi:13188\n  package already installed\n'
    )
    assert (
        run('install https://www.wowinterface.com/downloads/info13188-Molinari.html').output
        == '✗ wowi:13188\n  package already installed\n'
    )


def test_install_with_github_alias(run):
    assert run('install github:AdiAddons/AdiButtonAuras').output.startswith(
        '✓ github:AdiAddons/AdiButtonAuras\n  installed'
    )
    assert (
        run('install github:adiaddons/adibuttonauras').output
        == '✗ github:adiaddons/adibuttonauras\n  package already installed\n'
    )
    assert (
        run('install https://github.com/AdiAddons/AdiButtonAuras').output
        == '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'
    )
    assert (
        run('install https://github.com/AdiAddons/AdiButtonAuras/releases').output
        == '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'
    )
    assert (
        run('remove github:adiaddons/adibuttonauras').output
        == '✓ github:adiaddons/adibuttonauras\n  removed\n'
    )


def test_version_strategy_lifecycle(iw_config, run):
    assert run('install curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed 90100.79-Release'
    )
    assert (
        run('install --version foo curse:molinari').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is up to date\n'
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert (
        run('install --version foo curse:molinari').output
        == f"✗ curse:molinari\n  version foo not found\n"
    )
    assert (
        run('install --version 80000.57-Release curse:molinari').output
        == '✓ curse:molinari\n  installed 80000.57-Release\n'
    )
    assert run('update').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


def test_install_options(iw_config, run):
    assert run(
        'install'
        ' curse:molinari'
        ' -s latest curse:molinari'
        ' --version 80000.57-Release curse:molinari'
    ).output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_install_option_order_is_respected(run):
    assert run(
        'install'
        ' --version 80000.57-Release curse:molinari'
        ' -s latest curse:molinari'
        ' curse:molinari'
    ).output == dedent(
        '''\
        ✓ curse:molinari
          installed 80000.57-Release
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_install_argument_is_not_required(iw_config, run):
    assert run(
        'install -s latest curse:molinari --version 80000.57-Release curse:molinari'
    ).output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_configure__display_active_profile(iw_config, run):
    assert run('configure --active').output == iw_config.json(indent=2) + '\n'


def test_configure__create_new_profile(feed_pt, iw_config, run):
    feed_pt(f'{iw_config.addon_dir}\r\r')
    assert (
        run('-p foo configure').output
        == f'Configuration written to: {iw_config.config_dir / "profiles/foo/config.json"}\n'
    )


def test_configure__create_new_profile_promptless(monkeypatch, iw_config, run):
    monkeypatch.setenv('INSTAWOW_ADDON_DIR', str(iw_config.addon_dir))
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', iw_config.game_flavour.value)
    assert (
        run('-p foo configure --promptless').output
        == f'Configuration written to: {iw_config.config_dir / "profiles/foo/config.json"}\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__pkg_not_installed(run, options):
    assert (
        run(f'rollback {options} curse:molinari').output
        == '✗ curse:molinari\n  package is not installed\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__unsupported(run, options):
    assert run('install wowi:13188-molinari').exit_code == 0
    assert (
        run(f'rollback {options} wowi:13188-molinari').output
        == "✗ wowi:13188-molinari\n  'version' strategy is not valid for source\n"
    )


def test_rollback__single_version(run):
    assert run('install curse:molinari').exit_code == 0
    assert (
        run(f'rollback curse:molinari').output
        == '✗ curse:molinari\n  cannot find older versions\n'
    )


def test_rollback__multiple_versions(feed_pt, iw_config, run):
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    feed_pt('\r\r')
    assert run('rollback curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""} to 80000.57-Release
        '''
    )


def test_rollback__multiple_versions_promptless(iw_config, run):
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    assert run('rollback --version 80000.57-Release curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""} to 80000.57-Release
        '''
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__rollback_multiple_versions(feed_pt, iw_config, run, options):
    assert run('install curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    feed_pt('\r\r')
    assert run(f'rollback {options} curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 80000.57-Release to 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
        '''
    )


def test_rollback__cannot_use_version_with_undo(run):
    result = run(f'rollback --version foo --undo curse:molinari')
    assert result.exit_code == 2
    assert 'Cannot use "--version" and "--undo" together' in result.output


def test_reconcile__list_unreconciled(faux_molinari_and_run):
    assert faux_molinari_and_run('reconcile --list-unreconciled').output == (
        # fmt: off
        'unreconciled\n'
        '------------\n'
        'Molinari    \n'
        # fmt: on
    )


def test_reconcile__auto_reconcile(iw_config, faux_molinari_and_run):
    assert faux_molinari_and_run('reconcile --auto').output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
        '''
    )


def test_reconcile__abort_interactive_reconciliation(feed_pt, faux_molinari_and_run):
    feed_pt('\x03')  # ^C
    assert faux_molinari_and_run('reconcile').output.endswith('Aborted!\n')


def test_reconcile__complete_interactive_reconciliation(feed_pt, iw_config, faux_molinari_and_run):
    feed_pt('\r\r')
    assert faux_molinari_and_run('reconcile').output.endswith(
        dedent(
            f'''\
            ✓ curse:molinari
              installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
            '''
        )
    )


def test_reconcile__reconciliation_complete(run):
    assert run('reconcile').output == 'No add-ons left to reconcile.\n'


def test_search__no_results(feed_pt, run):
    assert run('search ∅').output == 'No results found.\n'


def test_search__exit_without_selecting(feed_pt, run):
    feed_pt('\r')  # enter
    assert run('search molinari').output == ''


def test_search__exit_after_selection(feed_pt, run):
    feed_pt(' \rn')  # space, enter, "n"
    assert run('search molinari').output == ''


def test_search__install_one(feed_pt, iw_config, run):
    feed_pt(' \r\r')  # space, enter, enter
    assert run('search molinari --source curse').output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
        '''
    )


def test_search__install_multiple_conflicting(feed_pt, iw_config, run):
    feed_pt(' \x1b[B \r\r')  # space, arrow down, space, enter, enter
    assert run('search molinari').output in {
        dedent(
            f'''\
            ✓ curse:molinari
              installed 90100.79-Release{"-classic" if iw_config.game_flavour is Flavour.vanilla_classic else ""}
            ✗ wowi:13188-molinari
              package folders conflict with installed package Molinari
                (curse:20338)
            '''
        ),
        dedent(
            f'''\
            ✓ wowi:13188-molinari
              installed 90100.79-Release
            ✗ curse:molinari
              package folders conflict with installed package Molinari
                (wowi:13188)
            '''
        ),
    }


def test_changelog_output_with_arg(molinari_and_run):
    assert molinari_and_run('view-changelog curse:molinari').output.startswith(
        '<h3>Changes in 90100.79-Release:</h3>'
    )


def test_argless_changelog_output(molinari_and_run):
    assert molinari_and_run('view-changelog').output.startswith(
        'curse:molinari:\n  <h3>Changes in 90100.79-Release:</h3>'
    )


@pytest.mark.parametrize(
    'command, exit_code',
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(monkeypatch, molinari_and_run, command, exit_code):
    monkeypatch.setattr('click.launch', lambda url, wait=..., locate=...: ...)
    assert molinari_and_run(command).exit_code == exit_code


def test_can_list_with_substr_match(molinari_and_run):
    assert molinari_and_run('list mol').output == 'curse:molinari\n'
    assert molinari_and_run('list foo').output == ''
    assert molinari_and_run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(molinari_and_run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export(iw_config, molinari_and_run):
    output = molinari_and_run('list -f json').output
    assert PkgList.parse_raw(output).__root__[0].name == 'Molinari'


def test_show_version(run):
    assert run('--version').output == f'instawow, version {__version__}\n'


def test_plugin_hook_command_can_be_invoked(run):
    pytest.importorskip('instawow_test_plugin')
    assert run('foo').output == 'success!\n'



import re

import pytest

from instawow import results as R
from instawow.common import Strategy
from instawow.config import Flavour
from instawow.models import Pkg
from instawow.resolvers import Defn


@pytest.mark.asyncio
@pytest.mark.parametrize('strategy', [Strategy.default, Strategy.latest, Strategy.any_flavour])
async def test_curse_common_strategies(iw_manager, request, strategy):
    retail_and_vanilla_classic_files = Defn('curse', 'molinari', strategy=strategy)
    retail_only_file = Defn('curse', 'mythic-dungeon-tools', strategy=strategy)
    classic_only_file = Defn('curse', 'classiccastbars', strategy=strategy)
    multiflavour_file = Defn('curse', 'elkbuffbars', strategy=strategy)

    results = await iw_manager.resolve(
        [retail_and_vanilla_classic_files, retail_only_file, classic_only_file, multiflavour_file]
    )
    if iw_manager.config.game_flavour is Flavour.vanilla_classic:
        if strategy is Strategy.any_flavour:
            assert type(results[retail_only_file]) is Pkg
        else:
            assert 'classic' in results[retail_and_vanilla_classic_files].version
            assert (
                type(results[retail_only_file]) is R.PkgFileUnavailable
                and results[retail_only_file].message
                == f"no files matching vanilla_classic using {strategy} strategy"
            )
        assert type(results[classic_only_file]) is Pkg
    elif iw_manager.config.game_flavour is Flavour.retail:
        assert type(results[retail_only_file]) is Pkg
        if strategy is Strategy.any_flavour:
            assert type(results[classic_only_file]) is Pkg
        else:
            assert 'classic' not in results[retail_and_vanilla_classic_files].version
            assert (
                type(results[classic_only_file]) is R.PkgFileUnavailable
                and results[classic_only_file].message
                == f"no files matching retail using {strategy} strategy"
            )


@pytest.mark.asyncio
async def test_curse_version_pinning(iw_manager):
    defn = Defn('curse', 'molinari').with_version('70300.51-Release')
    results = await iw_manager.resolve([defn])
    assert (
        results[defn].options.strategy == Strategy.version
        and results[defn].version == '70300.51-Release'
    )


@pytest.mark.asyncio
async def test_curse_deps_are_found(iw_manager):
    defn = Defn('curse', 'bigwigs-voice-korean')

    if iw_manager.config.game_flavour is not Flavour.retail:
        pytest.skip(f'{defn} is only available for retail')

    results = await iw_manager.resolve([defn], with_deps=True)
    assert ['bigwigs-voice-korean', 'big-wigs'] == [d.slug for d in results.values()]


@pytest.mark.asyncio
async def test_curse_changelog_is_url(iw_manager):
    molinari = Defn('curse', 'molinari')

    results = await iw_manager.resolve([molinari])
    assert re.match(
        r'https://addons-ecs\.forgesvc\.net/api/v2/addon/\d+/file/\d+/changelog',
        results[molinari].changelog_url,
    )


@pytest.mark.asyncio
async def test_wowi(iw_manager):
    retail_and_classic = Defn('wowi', '13188-molinari')
    unsupported_strategy = Defn('wowi', '13188', strategy=Strategy.latest)

    results = await iw_manager.resolve([retail_and_classic, unsupported_strategy])
    assert type(results[retail_and_classic]) is Pkg
    assert (
        type(results[unsupported_strategy]) is R.PkgStrategyUnsupported
        and results[unsupported_strategy].message == "'latest' strategy is not valid for source"
    )


@pytest.mark.asyncio
async def test_wowi_changelog_is_data_url(iw_manager):
    molinari = Defn('wowi', '13188-molinari')
    results = await iw_manager.resolve([molinari])
    assert results[molinari].changelog_url.startswith('data:,')


@pytest.mark.asyncio
async def test_tukui(iw_manager):
    regular_addon = Defn('tukui', '1')
    ui_suite = Defn('tukui', '-1')
    ui_suite_with_slug = Defn('tukui', 'tukui')
    unsupported_strategy = Defn('tukui', '1', strategy=Strategy.latest)

    results = await iw_manager.resolve(
        [regular_addon, ui_suite, ui_suite_with_slug, unsupported_strategy]
    )

    if iw_manager.config.game_flavour is Flavour.retail:
        assert (
            type(results[regular_addon]) is Pkg and results[regular_addon].name == 'MerathilisUI'
        )
        assert type(results[ui_suite]) is Pkg and results[ui_suite].name == 'Tukui'
        assert (
            type(results[ui_suite_with_slug]) is Pkg
            and results[ui_suite_with_slug].name == 'Tukui'
        )
    else:
        assert type(results[regular_addon]) is Pkg and results[regular_addon].name == 'Tukui'
        assert type(results[ui_suite]) is R.PkgNonexistent
        assert type(results[ui_suite_with_slug]) is R.PkgNonexistent
    assert (
        type(results[unsupported_strategy]) is R.PkgStrategyUnsupported
        and results[unsupported_strategy].message == "'latest' strategy is not valid for source"
    )


@pytest.mark.asyncio
async def test_tukui_changelog_url_varies_by_addon_type(iw_manager):
    ui_suite = Defn('tukui', '-1')
    regular_addon = Defn('tukui', '1')
    results = await iw_manager.resolve([ui_suite, regular_addon])
    if iw_manager.config.game_flavour is Flavour.retail:
        assert results[ui_suite].changelog_url == 'https://www.tukui.org/ui/tukui/changelog#20.17'
    assert results[regular_addon].changelog_url.startswith('data:,')


@pytest.mark.asyncio
async def test_github(iw_manager):
    release_json = Defn('github', 'nebularg/PackagerTest')
    legacy_lib_and_nolib = Defn('github', 'AdiAddons/AdiButtonAuras')
    legacy_latest = Defn('github', 'AdiAddons/AdiButtonAuras', strategy=Strategy.latest)
    legacy_older_version = Defn('github', 'AdiAddons/AdiButtonAuras').with_version('2.1.0')
    legacy_assetless = Defn('github', 'AdiAddons/AdiButtonAuras').with_version('2.0.19')
    legacy_retail_and_classic = Defn('github', 'p3lim-wow/Molinari')
    releaseless = Defn('github', 'AdiAddons/AdiBags')
    nonexistent = Defn('github', 'layday/foobar')

    results = await iw_manager.resolve(
        [
            release_json,
            legacy_lib_and_nolib,
            legacy_latest,
            legacy_older_version,
            legacy_assetless,
            legacy_retail_and_classic,
            releaseless,
            nonexistent,
        ]
    )
    assert ('classic' in results[release_json].download_url) is (
        iw_manager.config.game_flavour is Flavour.vanilla_classic
    )
    assert 'nolib' not in results[release_json].download_url
    assert 'nolib' not in results[legacy_lib_and_nolib].download_url
    assert (
        results[legacy_latest].options.strategy == Strategy.latest
        and 'nolib' not in results[legacy_latest].download_url
    )
    assert (
        results[legacy_older_version].options.strategy == Strategy.version
        and results[legacy_older_version].version == '2.1.0'
    )
    assert type(results[legacy_assetless]) is R.PkgFileUnavailable
    assert ('classic' in results[legacy_retail_and_classic].download_url) is (
        iw_manager.config.game_flavour is Flavour.vanilla_classic
    )
    assert (
        type(results[releaseless]) is R.PkgFileUnavailable
        and results[releaseless].message == 'release not found'
    )
    assert type(results[nonexistent]) is R.PkgNonexistent


@pytest.mark.asyncio
async def test_github_changelog_is_data_url(iw_manager):
    adibuttonauras = Defn('github', 'AdiAddons/AdiButtonAuras')
    results = await iw_manager.resolve([adibuttonauras])
    assert results[adibuttonauras].changelog_url.startswith('data:,')



from __future__ import annotations

import importlib.util
import sys


def __getattr__(name: str) -> object:
    """Defer importing own modules until attempting to access an attribute.

    Importing this in ``__init__`` will overwrite relative imports.
    """
    fullname = __package__ + '.' + name
    try:
        return sys.modules[fullname]
    except KeyError:
        # ``module_from_spec`` will raise if the spec is ``None`` but this is
        # converted to an ``ImportError`` by Python's import mechanism so we
        # don't have to be too fussy about what returns which - everything's
        # gonna work out just fine, I promise
        spec = importlib.util.find_spec(fullname)
        sys.modules[fullname] = module = importlib.util.module_from_spec(
            spec,  # type: ignore
        )
        lazy_loader = importlib.util.LazyLoader(
            spec.loader,  # type: ignore
        )
        lazy_loader.exec_module(module)
        return module



# pyright: reportGeneralTypeIssues=warning
# pyright: reportUnknownArgumentType=false

from __future__ import annotations

from functools import partial
import json
import platform

import toga
import toga.style

from . import json_rpc_server


class InstawowApp(toga.App):
    app: InstawowApp

    def __init__(self, **kwargs: object) -> None:
        super().__init__(
            formal_name='instawow-gui',
            app_id='org.instawow.instawow_gui',
            app_name='instawow_gui',
            icon='resources/instawow_gui',
            **kwargs,
        )

    async def _startup(self, native_app: object) -> None:
        server_url, serve = await json_rpc_server.prepare()
        self._iw_web_view.url = str(server_url)
        await serve()

    def startup(self) -> None:
        self.main_window = self.iw_window = toga.MainWindow(
            title=self.formal_name, size=(800, 600)
        )
        if platform.system() == 'Windows':
            from . import cef_adapter

            cef_adapter.load()

            self.iw_window.content = self._iw_web_view = toga.WebView(
                style=toga.style.Pack(flex=1), factory=cef_adapter.Factory
            )

        else:
            self.iw_window.content = self._iw_web_view = toga.WebView(
                style=toga.style.Pack(flex=1)
            )

        def dispatch_js_keyboard_event(_: toga.Command, *, action: str):
            event_args = json.dumps(
                {'detail': {'action': action}, 'bubbles': True, 'cancelable': True}
            )
            self._iw_web_view.invoke_javascript(
                f'document.dispatchEvent(new CustomEvent("togaSimulateKeypress", {event_args}));'
            )

        self.commands.add(
            toga.Command(
                partial(dispatch_js_keyboard_event, action='toggleSearchFilter'),
                label='Toggle Search Filter',
                shortcut=toga.Key.MOD_1 + toga.Key.G,
                group=toga.Group.EDIT,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewInstalled'),
                label='Installed',
                shortcut=toga.Key.MOD_1 + toga.Key.L,
                group=toga.Group.WINDOW,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewReconcile'),
                label='Unreconciled',
                group=toga.Group.WINDOW,
                section=20,
                order=20,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewSearch'),
                label='Search',
                shortcut=toga.Key.MOD_1 + toga.Key.F,
                group=toga.Group.WINDOW,
                section=20,
                order=30,
            ),
        )

        self.add_background_task(self._startup)

        self.iw_window.show()



from alembic import context
from sqlalchemy import engine_from_config, pool

from instawow.db import metadata as target_metadata


def run_offline() -> None:
    url = context.config.get_main_option('sqlalchemy.url')
    context.configure(
        url=url, target_metadata=target_metadata, render_as_batch=True, literal_binds=True
    )
    with context.begin_transaction():
        context.run_migrations()


def run_online() -> None:
    alembic_config = context.config.get_section(context.config.config_ini_section)
    if not alembic_config.get('sqlalchemy.url'):
        url = context.get_x_argument(as_dictionary=True).get('url')
        alembic_config['sqlalchemy.url'] = url

    engine = engine_from_config(alembic_config, prefix='sqlalchemy.', poolclass=pool.NullPool)
    with engine.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata, render_as_batch=True
        )
        with context.begin_transaction():
            context.run_migrations()


def main() -> None:
    if context.is_offline_mode():
        run_offline()
    else:
        run_online()


if __name__ == 'env_py':  # Alembic appears to mangle the module's `__name__`
    main()
